#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) 2022, Rafael Neto Henriques, Benjamin Ades-Aron
#
# Developers : Rafael Neto Henriques (rafaelnh21@gmail.com)
#              Benjamin Ades-Aron (benjamin.ades-aron@nyulangone.org)
# -------------------------------------------------------------------------
# Adapted implementation of the gibbs removal procedure suggested by:
#
# Kellner E, Dhital B, Kiselev VG, Reisert M. Gibbs-ringing artifact removal
# based on local subvoxel-shifts. Magn Reson Med. 2015 doi: 10.1002/mrm.26054.
#
# Lee, H.H., Novikov, D.S., Fieremans, E., 2021. Removal of partial 
# Fourier‐induced Gibbs (RPG) ringing artifacts in MRI. 
# Magn Reson in Med. 86(5), pp.2733-2750.
#
# Full adaption of the original fully-sampled code is described in chapter 3
# of thesis:
#
# Neto Henriques, R., 2018. Advanced Methods for Diffusion MRI Data
# Analysis and their Application to the Healthy Ageing Brain
# (Doctoral thesis). https://doi.org/10.17863/CAM.29356
# -------------------------------------------------------------------------

import numpy as np
from tqdm import tqdm
from functools import partial
from multiprocessing import Pool
import scipy.fft as fft
from joblib import Parallel, delayed

class RPG(object):

    def _image_tv(self, x, axis=0, n_points=3):
        """ 
        Computes total variation (TV) of matrix x accross a given axis and
        along two directions.

        Parameters
        ----------
        x : 2D ndarray
            matrix x
        axis : int (0 or 1)
            Axis which TV will be calculated. Default a is set to 0.
        n_points : int
            Number of points to be included in TV calculation.

        Returns
        -------
        ptv : 2D ndarray
            Total variation calculated from the right neighbours of each point
        ntv : 2D ndarray
            Total variation calculated from the left neighbours of each point
        """
        xs = x.copy() if axis else x.T.copy()

        # Add copies of the data so that data extreme points are also analysed
        xs = np.concatenate((xs[:, (-n_points-1):], xs, xs[:, 0:(n_points+1)]),
                            axis=1)

        ptv = np.absolute(xs[:, (n_points+1):(-n_points-1)] -
                        xs[:, (n_points+2):(-n_points)])
        ntv = np.absolute(xs[:, (n_points+1):(-n_points-1)] -
                        xs[:, (n_points):(-n_points-2)])
        for n in range(1, n_points):
            ptv = ptv + np.absolute(xs[:, (n_points+1+n):(-n_points-1+n)] -
                                    xs[:, (n_points+2+n):(-n_points+n)])
            ntv = ntv + np.absolute(xs[:, (n_points+1-n):(-n_points-1-n)] -
                                    xs[:, (n_points-n):(-n_points-2-n)])

        if axis:
            return ptv, ntv
        else:
            return ptv.T, ntv.T

    def nn_resample(self, img, shape):
        """
        Computes the nearest neighbor resampling of a 2D image to a new grid shape

        Parameters
        ----------
        img : 2D ndarray
        shape : tuple of ints 
            Dimensions of the new grid img will be sampled to

        Returns
        -------
        img : 2D ndarray
            New image sampled to updated grid shape
        """
        def per_axis(in_sz, out_sz):
            ratio = 0.5 * in_sz / out_sz
            return np.round(np.linspace(ratio - 0.5, in_sz - ratio - 0.5, num=out_sz)).astype(int)

        return img[per_axis(img.shape[0], shape[0])[:, None],
                per_axis(img.shape[1], shape[1])]


    def _gibbs_removal_1d(self, x, axis=0, n_points=3):
        """ 
        Suppresses Gibbs ringing along a given axis using fourier sub-shifts.

        Parameters
        ----------
        x : 2D ndarray
            Matrix x.
        axis : int (0 or 1)
            Axis in which Gibbs oscillations will be suppressed. Default is set
            to 0
        n_points : int, optional
            Number of neighbours to access local TV (see note). Default is set to
            3.

        Returns
        -------
        xc : 2D ndarray
            Matrix with suppressed Gibbs oscillations along the given axis.

        Note
        ----
        This function suppresses the effects of Gibbs oscillations based on the
        analysis of local total variation (TV). Although artefact correction is
        done based on two adjanced points for each voxel, total variation should be
        accessed in a larger range of neighbours. The number of the neighbours to
        be considered in TV calculation can be adjusted using parameter n_points.
        """
        ssamp = np.linspace(0.01, .6, num=41)

        xs = x.copy() if axis else x.T.copy()
        
        # TV for shift zero (baseline)
        tvr, tvl = self._image_tv(xs, axis=1, n_points=n_points)
        tvp = np.minimum(tvr, tvl)
        tvn = tvp.copy()

        # Find optimal shift for gibbs removal
        isp = xs.copy()
        isn = xs.copy()
        sp = np.zeros(xs.shape)
        sn = np.zeros(xs.shape)
        N = xs.shape[1]
        c = fft.fft(xs, axis=1, norm='ortho', n=N)
        k = fft.fftfreq(N, 1 / (2.0j * np.pi))
        for s in ssamp:
            ks = k * s
            # Access positive shift for given s
            img_p = abs(fft.ifft(c * np.exp(ks), axis=1, norm='ortho', n=N))
            tvsr, tvsl = self._image_tv(img_p, axis=1, n_points=n_points)
            tvs_p = np.minimum(tvsr, tvsl)

            # Access negative shift for given s
            img_n = abs(fft.ifft(c * np.exp(-ks), axis=1, norm='ortho', n=N))
            tvsr, tvsl = self._image_tv(img_n, axis=1, n_points=n_points)
            tvs_n = np.minimum(tvsr, tvsl)

            # Update positive shift params
            isp[tvp > tvs_p] = img_p[tvp > tvs_p]
            sp[tvp > tvs_p] = s
            tvp[tvp > tvs_p] = tvs_p[tvp > tvs_p]

            # Update negative shift params
            isn[tvn > tvs_n] = img_n[tvn > tvs_n]
            sn[tvn > tvs_n] = s
            tvn[tvn > tvs_n] = tvs_n[tvn > tvs_n]

        # check non-zero sub-voxel shifts
        idx = np.nonzero(sp + sn)

        # use positive and negative optimal sub-voxel shifts to interpolate to
        # original grid points

        xs[idx] = (isp[idx] - isn[idx])/(sp[idx] + sn[idx])*sn[idx] + isn[idx]

        # import scipy.interpolate as sci
        # import matplotlib.pyplot as plt
        # plt.imshow(img_p); plt.colorbar(); plt.show(); 
        # # #xs_ = sci.interp2d(idx[0], idx[1], vals[idx], kind='cubic')
        # import pdb;
        # pdb.set_trace()

        return xs if axis else xs.T


    def _weights(self, shape):
        """ 
        Computes the weights necessary to combine two images processed by
        the 1D Gibbs removal procedure along two different axes [1]_.

        Parameters
        ----------
        shape : tuple
            shape of the image

        Returns
        -------
        G0 : 2D ndarray
            Weights for the image corrected along axis 0.
        G1 : 2D ndarray
            Weights for the image corrected along axis 1.

        References
        ----------
        .. [1] Kellner E, Dhital B, Kiselev VG, Reisert M. Gibbs-ringing artifact
            removal based on local subvoxel-shifts. Magn Reson Med. 2016
            doi: 10.1002/mrm.26054.
        """

        G0 = np.zeros(shape)
        G1 = np.zeros(shape)
        k0 = np.linspace(-np.pi, np.pi, num=shape[0])
        k1 = np.linspace(-np.pi, np.pi, num=shape[1])

        # Middle points
        K1, K0 = np.meshgrid(k1[1:-1], k0[1:-1])
        cosk0 = 1.0 + np.cos(K0)
        cosk1 = 1.0 + np.cos(K1)
        G1[1:-1, 1:-1] = cosk0 / (cosk0+cosk1)
        G0[1:-1, 1:-1] = cosk1 / (cosk0+cosk1)

        # Boundaries
        G1[1:-1, 0] = G1[1:-1, -1] = 1
        G1[0, 0] = G1[-1, -1] = G1[0, -1] = G1[-1, 0] = 1/2
        G0[0, 1:-1] = G0[-1, 1:-1] = 1
        G0[0, 0] = G0[-1, -1] = G0[0, -1] = G0[-1, 0] = 1/2

        return G0, G1


    def _gibbs_removal_2d(self, image, n_points=3, G0=None, G1=None):
        """ 
        Suppress Gibbs ringing of a 2D image.

        Parameters
        ----------
        image : 2D ndarray
            Matrix cotaining the 2D image.
        n_points : int, optional
            Number of neighbours to access local TV (see note). Default is
            set to 3.
        G0 : 2D ndarray, optional.
            Weights for the image corrected along axis 0. If not given, the
            function estimates them using function :func:`_weights`
        G1 : 2D ndarray
            Weights for the image corrected along axis 1. If not given, the
            function estimates them using function :func:`_weights`

        Returns
        -------
        imagec : 2D ndarray
            Matrix with Gibbs oscillations reduced along axis a.

        References
        ----------
        Please cite the following articles
        .. [1] Neto Henriques, R., 2018. Advanced Methods for Diffusion MRI Data
            Analysis and their Application to the Healthy Ageing Brain
            (Doctoral thesis). https://doi.org/10.17863/CAM.29356
        .. [2] Kellner E, Dhital B, Kiselev VG, Reisert M. Gibbs-ringing artifact
            removal based on local subvoxel-shifts. Magn Reson Med. 2016
            doi: 10.1002/mrm.26054.
        """
        if np.any(G0) is None or np.any(G1) is None:
            G0, G1 = self._weights(image.shape)

        img_c1 = self._gibbs_removal_1d(image, axis=1, n_points=n_points)
        img_c0 = self._gibbs_removal_1d(image, axis=0, n_points=n_points)

        C1 = fft.fft2(img_c1, norm='ortho')
        C0 = fft.fft2(img_c0, norm='ortho')
        imagec = abs(fft.ifft2(fft.fftshift(C1)*G1 + fft.fftshift(C0)*G0, norm='ortho'))
       
        return imagec

    def _gibbs_removal_2d_ff(self, image, n_points=3, G0=None, G1=None):
        if np.any(G0) is None or np.any(G1) is None:
            G0, G1 = self._weights(image.shape)

        img_f1 = abs(fft.ifft2(fft.fftshift(fft.fft2(image, norm='ortho'))*G1, norm='ortho'))
        img_f0 = abs(fft.ifft2(fft.fftshift(fft.fft2(image, norm='ortho'))*G0, norm='ortho'))

        img_c1 = self._gibbs_removal_1d(img_f1, axis=1, n_points=n_points)
        img_c0 = self._gibbs_removal_1d(img_f0, axis=0, n_points=n_points)

        imagec = abs(img_c1 + img_c0)

        return imagec

    def _gibbs_removal_2d_y(self, image, n_points=3):
        """ 
        Suppress Gibbs ringing of a 2D image along a single dimension.
        For use with partial fourier factors 5/8 and 7/8.

        Parameters
        ----------
        image : 2D ndarray
            Matrix cotaining the 2D image.
        n_points : int, optional
            Number of neighbours to access local TV (see note). Default is
            set to 3.

        Returns
        -------
        imagec : 2D ndarray
            Matrix with Gibbs oscillations reduced only in 1 direction.

        References
        ----------
        Please cite the following articles
        .. [1] Neto Henriques, R., 2018. Advanced Methods for Diffusion MRI Data
            Analysis and their Application to the Healthy Ageing Brain
            (Doctoral thesis). https://doi.org/10.17863/CAM.29356
        .. [2] Kellner E, Dhital B, Kiselev VG, Reisert M. Gibbs-ringing artifact
            removal based on local subvoxel-shifts. Magn Reson Med. 2016
            doi: 10.1002/mrm.26054.
        .. [3] Lee, H.H., Novikov, D.S., Fieremans, E., 2021. Removal of partial 
            Fourier-induced Gibbs (RPG) ringing artifacts in MRI. 
            Magn Reson in Med. 86(5), pp.2733-2750.
        """

        # remove gibbs along axis 1
        imagec = self._gibbs_removal_1d(image, axis=0, n_points=n_points)

        return imagec

    def _gibbs_removal_2d_y_2(self, image, n_points=3, G0=None, G1=None):
        """ 
        Suppress Gibbs ringing of a 2D image with 6/8 partial fourier.

        Parameters
        ----------
        image : 2D ndarray
            Matrix cotaining the 2D image.
        n_points : int, optional
            Number of neighbours to access local TV (see note). Default is
            set to 3.
        G0 : 2D ndarray, optional.
            Weights for the image corrected along axis 0. If not given, the
            function estimates them using function :func:`_weights`
        G1 : 2D ndarray
            Weights for the image corrected along axis 1. If not given, the
            function estimates them using function :func:`_weights`

        Returns
        -------
        imagec : 2D ndarray
            Matrix with Gibbs oscillations reduced along axis a.

        References
        ----------
        Please cite the following articles
        .. [1] Neto Henriques, R., 2018. Advanced Methods for Diffusion MRI Data
            Analysis and their Application to the Healthy Ageing Brain
            (Doctoral thesis). https://doi.org/10.17863/CAM.29356
        .. [2] Kellner E, Dhital B, Kiselev VG, Reisert M. Gibbs-ringing artifact
            removal based on local subvoxel-shifts. Magn Reson Med. 2016
            doi: 10.1002/mrm.26054.
        .. [3] Lee, H.H., Novikov, D.S., Fieremans, E., 2021. Removal of partial 
            Fourier-induced Gibbs (RPG) ringing artifacts in MRI. 
            Magn Reson in Med. 86(5), pp.2733-2750.
        """
        if np.any(G0) is None or np.any(G1) is None:
            G0, G1 = self._weights(image.shape)

        img_c1 = self._gibbs_removal_1d(image, axis=1, n_points=n_points)
        img_c0 = self._gibbs_removal_1d(image, axis=0, n_points=n_points)

        img_c00 = img_c0[::2,:]
        img_c01 = img_c0[1::2,:]

        img_c10 = self._gibbs_removal_1d(img_c00, axis=0, n_points=n_points)
        img_c11 = self._gibbs_removal_1d(img_c01, axis=0, n_points=n_points)
        img_c = np.zeros_like(image)

        img_c[::2,:] = img_c10
        img_c[1::2,:] = img_c11

        C1 = fft.fft2(img_c)
        C0 = fft.fft2(img_c1)
        imagec = abs(fft.ifft2(fft.fftshift(C1)*G1 + fft.fftshift(C0)*G0))

        return imagec

    def _parallel_pf_gibbs_removal(self, vol, shap, pf_fact, n_points, G0, G1, testff):
        """ 
        Wrapper function for parallel processing
        """

        if pf_fact == 5/8:
            scale = 4
            for sc in range(scale):
                vol[sc::scale,:] = self._gibbs_removal_2d(vol[sc::scale,:], n_points=n_points)
            vol = self._gibbs_removal_2d_y(vol, n_points=n_points)

        elif pf_fact == 6/8:
            vol = self._gibbs_removal_2d_y_2(vol, n_points=n_points)
            #vol = self._gibbs_removal_2d(vol, n_points=n_points, G0=G0, G1=G1)
            # from hong-hsi's supplemental material, the following will have better perfamnce
            # but oversmooth:
            # scale = 2
            # for sc in range(scale):
            #     vol[sc::scale,:] = _gibbs_removal_2d(vol[sc::scale,:], n_points=n_points)
        
        elif pf_fact == 7/8:
            scale = 4
            vol = self.nn_resample(vol, (shap[1]*3, shap[2]))
            for sc in range(scale):
                vol[sc::scale,:] = self._gibbs_removal_2d(vol[sc::scale,:], n_points=n_points)
            vol = self.nn_resample(vol, (shap[1], shap[2]))
            vol = self._gibbs_removal_2d_y(vol, n_points=n_points)

        
        elif pf_fact == 1 and testff:
            vol = self._gibbs_removal_2d_ff(vol, n_points=n_points, G0=G0, G1=G1)
        elif pf_fact == 1 and not testff:
            vol = self._gibbs_removal_2d(vol, n_points=n_points, G0=G0, G1=G1)

        return vol

    def gibbs_removal(self, vol, slice_axis=2, n_points=3, pe_dim=1, pf_fact=1, nproc=1, testff=False):
        """ 
        Suppresses Gibbs ringing artefacts of images volumes.

        Parameters
        ----------
        vol : ndarray ([X, Y]), ([X, Y, Z]) or ([X, Y, Z, g])
            Matrix containing one volume (3D) or multiple (4D) volumes of images.
        slice_axis : int (0, 1, or 2)
            Data axis corresponding to the number of acquired slices. Default is
            set to the third axis
        n_points : int, optional
            Number of neighbour points to access local TV (see note). Default is
            set to 3.

        Returns
        -------
        vol : ndarray ([X, Y]), ([X, Y, Z]) or ([X, Y, Z, g])
            Matrix containing one volume (3D) or multiple (4D) volumes of corrected
            images.

        Notes
        -----
        For 4D matrix last element should always correspond to the number of
        diffusion gradient directions.

        References
        ----------
        Please cite the following articles
        .. [1] Neto Henriques, R., 2018. Advanced Methods for Diffusion MRI Data
            Analysis and their Application to the Healthy Ageing Brain
            (Doctoral thesis). https://doi.org/10.17863/CAM.29356
        .. [2] Kellner E, Dhital B, Kiselev VG, Reisert M. Gibbs-ringing artifact
            removal based on local subvoxel-shifts. Magn Reson Med. 2016
            doi: 10.1002/mrm.26054.
        .. [3] Lee, H.H., Novikov, D.S., Fieremans, E., 2021. Removal of partial 
            Fourier-induced Gibbs (RPG) ringing artifacts in MRI. 
            Magn Reson in Med. 86(5), pp.2733-2750.
        """
        nd = vol.ndim
        
        # check the axis corresponding to different slices
        # 1) This axis cannot be larger than 2
        if slice_axis > 2:
            raise ValueError("Different slices have to be organized along" +
                            "one of the 3 first matrix dimensions")
        
        # check matrix dimension
        elif nd == 2:
            vol = vol[None,...]
        elif nd == 3:
            vol = np.moveaxis(vol, slice_axis, 0)
        elif nd == 4:
            vol = np.moveaxis(vol, (slice_axis, 3), (0,1))

        if nd == 4:
            inishap = vol.shape
            vol = vol.reshape((inishap[0] * inishap[1], inishap[2],  inishap[3]))
        
        if nd > 4:
            raise ValueError("Data have to be a 4D, 3D or 2D matrix")
        elif nd < 2:
            raise ValueError("Data is not an image")

        # Produce weigthing functions for 2D Gibbs removal
        shap = vol.shape
        G0, G1 = self._weights(shap[1:])

        # check that phase encoding axis is either 0 or 1
        if pe_dim > 2:
            raise ValueError("Phase encoding axis must be either 0 (LR) or 1 (AP)")

        # if phase incoding direction is along axis 0, swap first two axes
        if pe_dim == 1:
            vol = np.swapaxes(vol, 1, 2)

        pf_opts = [5/8, 6/8, 7/8, 1]
        if not pf_fact in pf_opts:
            raise ValueError("RPG degibbs only supports partial fourier factors of 5/8, 6/8, 7/8, and 1")

        vol = vol.copy()

        inputs = tqdm(range(vol.shape[0]))

        vol = (Parallel(n_jobs=nproc, prefer='processes')
            (delayed(self._parallel_pf_gibbs_removal)(
                vol[i,...], shap=shap, pf_fact=pf_fact, n_points=n_points, G0=G0, G1=G1, testff=testff
            ) for i in inputs))
        vol = np.array(vol)

        # pool = Pool(nproc)
        # partial_func = partial(
        #     self._parallel_pf_gibbs_removal, 
        #     shap=shap, pf_fact=pf_fact, n_points=n_points, G0=G0, G1=G1, testff=testff
        # )
        # vol[:, :, :] = pool.map(partial_func, vol)
        # pool.close()
        # pool.join()
    
        #vol[:, :, :] = self._parallel_pf_gibbs_removal(vol[0, :, :], shap=shap, pf_fact=pf_fact, n_points=n_points, G0=G0, G1=G1, testff=testff)

        # Reshape data to original format
        if pe_dim == 1:
            vol = np.swapaxes(vol, 1, 2)
        if nd == 2:
            vol = vol.squeeze()
        if nd == 3:
            vol = np.moveaxis(vol, 0, slice_axis)
        if nd == 4:
            vol = vol.reshape(inishap)
            vol = np.moveaxis(vol, (0, 1), (slice_axis, 3))

        return vol

def gibbs_removal(image, slice_axis=2, n_points=3, pe_dim=1, pf_fact=1, nproc=1, testff=False):
    rpg = RPG()
    return rpg.gibbs_removal(image, slice_axis, n_points, pe_dim, pf_fact, nproc, testff)


def fft2_mri(img):
    n = img.shape[0]*img.shape[1]
    return np.fft.fftshift(np.fft.fft2(np.fft.fftshift(img))) / 2

def ifft2_mri(img):
    n = img.shape[0]*img.shape[1]
    return np.fft.fftshift(np.fft.ifft2(np.fft.fftshift(img))) / 2

def make_gibbs_phantom(img, size, scale, pf, sigma, pe_dir=1):
    nps = np.ceil(size/scale)
    pk = fft2_mri(img)
    scale = (1-scale)/2
    pk = pk[round(img.shape[0]*scale):round(img.shape[0]*(1-scale)), 
            round(img.shape[1]*scale):round(img.shape[1]*(1-scale))]
    pk = pk[:size,:size]
    pk = pk + sigma * np.random.randn(size,size) + 1j * sigma * np.random.randn(size,size)
    p = ifft2_mri(pk)
    if pe_dir == 1:
        pk[:,round(size*pf):] = 0
    elif pe_dir == 0:
        pk[round(size*pf):,:] = 0
    p_pf = ifft2_mri(pk)
    return p, p_pf

if __name__ == '__main__':
    from phantominator import shepp_logan
    import matplotlib.pyplot as plt
    import scipy.io as sio

    sz = 128
    scale = 0.5
    pf = 6/8
    sigma = 0.02
    phantom = np.flip(shepp_logan(256))

    phantom_gibbs, phantom_gibbs_pf = make_gibbs_phantom(phantom, sz, scale, pf, sigma)
    sio.savemat('/Users/benaron/Desktop/gibbs_phantom.mat', {'phantom': phantom_gibbs})

    testff = True
    g1 = gibbs_removal(abs(phantom_gibbs), testff=testff, n_points=10)
    
    testff = False
    g2 = gibbs_removal(abs(phantom_gibbs), testff=testff, n_points=10)

    mat = sio.loadmat('/Volumes/Research/cbi05data/data1/Hamster/Ben/gibbs_phantom_hhdg.mat')
    g3 = mat['P_pf_dg']

    plt.subplot(3,3,1)
    plt.imshow(abs(phantom_gibbs), vmin=0, vmax=1)
    plt.title('gibbs phantom (no pf)'); plt.axis('off')
    plt.subplot(3,3,2)
    plt.imshow(abs(g1), vmin=0, vmax=1)
    plt.title('gibbs corrected (filter before unring1d)'); plt.axis('off')
    plt.subplot(3,3,3)
    plt.imshow(abs(g1)-abs(phantom_gibbs), vmin=-.25, vmax=.25)
    plt.title('residual'); plt.axis('off')
    plt.subplot(3,3,4)
    plt.imshow(abs(phantom_gibbs), vmin=0, vmax=1)
    plt.title('gibbs phantom (no pf)'); plt.axis('off')
    plt.subplot(3,3,5)
    plt.imshow(abs(g2), vmin=0, vmax=1)
    plt.title('gibbs corrected (filter after unring1d)'); plt.axis('off')
    plt.subplot(3,3,6)
    plt.imshow(abs(g2)-abs(phantom_gibbs), vmin=-.25, vmax=.25)
    plt.title('residual'); plt.axis('off')
    plt.subplot(3,3,7)
    plt.imshow(abs(phantom_gibbs), vmin=0, vmax=1)
    plt.title('gibbs phantom (no pf)'); plt.axis('off')
    plt.subplot(3,3,8)
    plt.imshow(g3, vmin=0, vmax=1)
    plt.title('gibbs corrected (Lee matlab code)'); plt.axis('off')
    plt.subplot(3,3,9)
    plt.imshow(abs(g3)-abs(phantom_gibbs), vmin=-.25, vmax=.25)
    plt.title('residual'); plt.axis('off')
    plt.show()






"""
Functions to run designer preprocessing steps

Includes:
run_denoising:

run_degibbs:

run_eddy: 

run_b1correct:

create_brainmask:

run_rice_bias_correct:

run_normalization:
"""

def run_mppca(args_extent, args_phase, args_shrinkage, args_algorithm):
    """
    wrapper for complex or magnitude, adatpive or local mppca
    
    """
    from mrtrix3 import run, app
    from ants import image_read, image_write, from_numpy
    import numpy as np

    if app.ARGS.adaptive_patch:
        import lib.mpdenoise_sj as mp
    else:
        import lib.mpcomplex as mp

    if args_extent:
        extent = args_extent
        if type(extent) == str:
            extent = extent.split(",")
            extent = [int(i) for i in extent]
    else: 
        extent = [5,5,5]

    run.command('mrconvert dwi.mif -export_grad_mrtrix grad.txt tmp_dwi.nii', show=False)
    nii = image_read('tmp_dwi.nii')
    dwi = nii.numpy()

    # note adaptive patch does not include mpcomplex
    n_cores = app.ARGS.n_cores
    print('...denoising...')
    if app.ARGS.adaptive_patch:
        Signal, Sigma, Nparameters = mp.denoise(
            dwi, patchtype='nonlocal', shrinkage=args_shrinkage, algorithm=args_algorithm)
        Sigma[np.isnan(Sigma)] = 0
    else:    
        Signal, Sigma, Nparameters = mp.denoise(
            dwi, phase=args_phase, kernel=extent, shrinkage=args_shrinkage, algorithm=args_algorithm, n_cores=n_cores
            )
    Sigma[np.isnan(Sigma)] = 0

    out = from_numpy(
        Signal, origin=nii.origin, spacing=nii.spacing, direction=nii.direction)
    image_write(out, 'tmp_dwidn.nii')
    out = from_numpy(
        Sigma, origin=nii.origin[:-1], spacing=nii.spacing[:-1], direction=nii.direction[:-1,:])
    image_write(out, 'sigma.nii')
    out = from_numpy(
        Nparameters, origin=nii.origin[:-1], spacing=nii.spacing[:-1], direction=nii.direction[:-1,:])
    image_write(out, 'Npars.nii')

    run.command('mrconvert -grad grad.txt tmp_dwidn.nii dwidn.mif', show=False)
    run.command('mrconvert sigma.nii noisemap.mif', show=False)
    app.cleanup('tmp_dwi.nii')
    app.cleanup('tmp_dwidn.nii')
    app.cleanup('grad.txt')
    
    #run.command('dwidenoise -noise fullnoisemap.mif -estimator Exp2 working.mif dwidn.mif')
    run.command('mrconvert -force dwidn.mif working.mif', show=False)

def run_patch2self():
    """
    wrapper for patch2self
    """
    from mrtrix3 import run, app
    from dipy.denoise.patch2self import patch2self
    from ants import image_read, image_write, from_numpy
    import numpy as np

    run.command('mrconvert -force -export_grad_fsl working.bvec working.bval working.mif working.nii', show=False)
    nii = image_read('working.nii')
    dwi = nii.numpy()

    bvals = np.loadtxt('working.bval')
    dwi_dn = patch2self(dwi, bvals)
    out = from_numpy(
        dwi_dn, origin=nii.origin, spacing=nii.spacing, direction=nii.direction)
    image_write(out, 'working_p2s.nii')
    run.command('mrconvert -force -fslgrad working.bvec working.bval working_p2s.nii working.mif', show=False)


def run_degibbs(pf, pe_dir):
    """
    wrapper for rpg degibbs
    """

    import lib.gibbs_removal_rpg as rpg
    from mrtrix3 import run, app
    from ants import image_read, image_write, from_numpy
    import os

    rpg_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))),'rpg_cpp')

    # convert working.mif to nii
    run.command('mrconvert -force -export_grad_fsl working.bvec working.bval working.mif working.nii', show=False)
    nii = image_read('working.nii')
    dwi = nii.numpy()
    print("...RPG degibbsing...")

    if pe_dir[0] == 'i':
        pe_rpg = 0
    elif pe_dir[0] == 'j':
        pe_rpg = 1
    elif pe_dir[0] == 'k':
        pe_rpg = 2
    
    n_cores = app.ARGS.n_cores

    print('Gibbs correction with parameters:')
    print('phase-encoding direction = %s' % (pe_dir))
    print('partial-fourier factor   = %s' % (pf))


    if 'i' in pe_dir:
        pe_dir = 'x'
    elif 'j' in pe_dir:
        pe_dir = 'y'
    elif 'k' in pe_dir:
        pe_dir = 'z'
    else:
        pe_dir = 'y'

    run.command('%s -pf %s -dim %s %s %s' %
                (os.path.join(rpg_path,'rpg'),
                pf, 
                pe_dir,
                'working.nii',
                'working_rpg.nii'
                ), show=False)

    # img_dg = rpg.gibbs_removal(dwi, pf_fact=pf, pe_dim=pe_rpg, nproc=n_cores)
    # out = from_numpy(
    #     img_dg, origin=nii.origin, spacing=nii.spacing, direction=nii.direction)
    # image_write(out, 'working_rpg.nii')

    #convert gibbs corrected nii to .mif
    run.command('mrconvert -force -fslgrad working.bvec working.bval working_rpg.nii working.mif', show=False)


def group_alignment(group_list):
    from mrtrix3 import app, run, path, image, fsl, MRtrixError
    import numpy as np

    fsl_suffix = fsl.suffix()

    print('... Rigidly aligning groups...')

    # 1) find the longest series and extract the b0
    group_len = []
    for i in range(len(group_list)):
        std = run.command('mrinfo -size %s' %
                          (path.to_scratch(group_list[i])
                           ), show=False)
        group_len.append([int(j) for j in std[0].strip().split(' ')][-1])
    longest_series = np.argmax(group_len)

    # idxlist = dwi_metadata['idxlist']
    # series_inds = [i.split(",") for i in idxlist]
    # series_lens = [len(i) for i in series_inds]
    # longest_series = np.argmax(series_lens)

    # For each group: Register b0 ref PE and b0 RPE to the b0 and run topup + eddy
    for i in range(len(group_list)):
        run.command('dwiextract -force -bzero %s - | mrmath - mean %s -axis 3 -force' %
                    (path.to_scratch(group_list[i]), 
                    path.to_scratch('b0_align_series_' + str(i) + '.nii')),
                    show=False)
        run.command('bet %s %s -f 0.2 -m' %
                    (path.to_scratch('b0_align_series_' + str(i) + '.nii'), 
                    path.to_scratch('b0_align_series_' + str(i) + '_brain')),
                    show=False)

    # 2) extract b0 images from all other series
    for i in range(len(group_list)):                
        if i == longest_series:
            continue
        else:
            # 3) rigidly register all b0 images to the b0 from the longest series
            run.command('flirt -in %s -ref %s -omat %s -dof 6' %
                    (path.to_scratch('b0_align_series_' + str(i) + '_brain'),
                    path.to_scratch('b0_align_series_' + str(longest_series) + '_brain'),
                    path.to_scratch('b0_align_series_to_longest_series_' + str(i) + '.mat')),
                    show=False)
            run.command('transformconvert -force %s %s %s flirt_import %s' %
                (path.to_scratch('b0_align_series_to_longest_series_' + str(i) + '.mat'),
                path.to_scratch('b0_align_series_' + str(i) + '_brain' + fsl_suffix),
                path.to_scratch('b0_align_series_' + str(longest_series) + '_brain' + fsl_suffix),
                path.to_scratch('b0_align_series_scan' + str(i) + 'to_longest_series_mrtrix.txt')),
                show=False)
            run.command('mrtransform -force -linear %s -interp cubic %s %s' %
                (path.to_scratch('b0_align_series_scan' + str(i) + 'to_longest_series_mrtrix.txt'),
                path.to_scratch(group_list[i]),
                path.to_scratch('dwi_align_series_' + str(i) + '_to_ref.mif')),
                show=False)
    
    # concatenate data in correct order
    series_cat_all = ''
    for i in range(len(group_list)):
        if i == longest_series:
            series_to_cat = path.to_scratch(group_list[i])
        else:
            series_to_cat = path.to_scratch('dwi_align_series_' + str(i) + '_to_ref.mif')
        series_cat_all = series_cat_all + series_to_cat + ' '
    run.command('mrcat -force %s %s' %
                (series_cat_all,
                path.to_scratch('dwi_align_series_aligned.mif')),
                show=False)
    run.command('mrconvert -force dwi_align_series_aligned.mif working.mif', show=False)


def convert_ants_xform(mat, i):
    import scipy.io as sio
    import numpy as np
    from mrtrix3 import app, run, path, image, fsl, MRtrixError

    ants_dict = sio.loadmat(mat)
    lps2ras = np.diag([-1, -1, 1])
    rot = ants_dict['AffineTransform_float_3_3'][0:9].reshape((3, 3))
    trans = ants_dict['AffineTransform_float_3_3'][9:12]
    offset = ants_dict['fixed']
    r_trans = (np.dot(rot, offset) - offset - trans).T * [1, 1, -1]
    
    data = np.eye(4)
    data[0:3, 3] = r_trans
    data[:3, :3] = np.dot(np.dot(lps2ras, rot), lps2ras)

    #np.savetxt('ants_textfile_' + str(i) + '_np.txt', data, fmt='%10.10f', delimiter=' ')
    return data

def pre_eddy_ants_moco(dwi_metadata):
    import ants
    from mrtrix3 import app, run, path, image, fsl, MRtrixError
    import numpy as np
    
    run.command('mrconvert -force -export_grad_fsl working.bvec working.bval working.mif working.nii', show=False)
    nii = ants.image_read('working.nii')
    ants_moco_params = ants.motion_correction(nii, moreaccurate=2)
    ants.image_write(ants_moco_params['motion_corrected'], 'working_antsmoco.nii')

    # ants_interp = ants.apply_transforms(fixed=nii, moving=nii, 
    #                                     transformlist=ants_moco_params['motion_parameters'],
    #                                     interpolator='bSpline', imagetype=3)

    dirs = dwi_metadata['grad'][:,:3]
    dirs_rot = np.zeros_like(dirs)
    # rotate gradients
    #img_cat_all = ''
    for i in range(dwi_metadata['grad'].shape[0]):

        antsaffine = ants_moco_params['motion_parameters'][i][0]
        aff = convert_ants_xform(antsaffine, i)
        diri = np.hstack((dirs[i,:],0))
        dirs_rot[i,:] = (aff @ diri.T)[:3]

    # run.command('mrconvert working.mif -coord 3 %s -axes 0,1,2 dwi_antsmoco_%s.mif' % 
    #             (str(i), str(i)), show=False
    #             )
    #     run.command('mrtransform -linear %s %s %s' %
    #                 ('ants_textfile_' + str(i) + '_np.txt',
    #                 'dwi_antsmoco_' + str(i) + '.mif',
    #                 'dwi_antsmoco_' + str(i) + '_withgrad.mif'),
    #                 show=False)
    #     series_to_cat = path.to_scratch('dwi_antsmoco_' + str(i) + '_withgrad.mif')
    #     img_cat_all = img_cat_all + series_to_cat + ' '
    # run.command('mrcat %s %s' %
    #             (img_cat_all,
    #             path.to_scratch('dwi_pre_eddy_antsmoco.mif')))

    np.savetxt('working_antsmoco.bvec', dirs_rot.T, delimiter=' ', fmt='%4.10f')
    run.command('mrconvert -force -fslgrad working_antsmoco.bvec working.bval working_antsmoco.nii working.mif', show=False)

def run_pre_align(dwi_metadata):
    from mrtrix3 import app, run, path, image, fsl, MRtrixError

    if len(dwi_metadata['dwi_list']) == 1:
        print('Warning -pre_align option was used but only 1 input series was provided')
    else:
        series_list = []
        for i in range(len(dwi_metadata['dwi_list'])):
            run.command('mrconvert -coord 3 %s working.mif %s' % 
                    (dwi_metadata['idxlist'][i], path.to_scratch('dwi_align_series_' + str(i) + '.mif')),
                    show=False)
            series_list.append('dwi_align_series_' + str(i) + '.mif')
            
        group_alignment(series_list)

def run_ants_moco(dwi_metadata):
    from mrtrix3 import app, run, path, image, fsl, MRtrixError

    pre_eddy_ants_moco(dwi_metadata)

def run_eddy(shell_table, dwi_metadata):
    from mrtrix3 import app, run, path, image, fsl, MRtrixError
    from lib.designer_input_utils import splitext_
    import numpy as np
    import json
    import os
    import glob

    print("...Eddy current, EPI, motion correction...")

    eddyopts = '" --cnr_maps --repol --data_is_shelled "'

    fsl_suffix = fsl.suffix()

    pe_dir = dwi_metadata['pe_dir']

    echo_times_per_series = dwi_metadata['TE']
    bshape_per_series = dwi_metadata['bshape']
    if  echo_times_per_series.count(echo_times_per_series[0]) == len(echo_times_per_series):
        flag_variable_TE = False
    else:
        flag_variable_TE = True

    if  bshape_per_series.count(bshape_per_series[0]) == len(bshape_per_series):
        flag_variable_bshape = False
    else:
        flag_variable_bshape = True

    if flag_variable_TE or flag_variable_bshape:
        if app.ARGS.eddy_groups is None:
            raise MRtrixError('for variable TE data, the user must supply the \
                            -eddy_groups command line argument')
        
         # get TE of the PA
        if app.ARGS.rpe_pair:
            if app.ARGS.rpe_te:
                te_pa = float(app.ARGS.rpe_te)
            else:
                pa_fpath = splitext_(path.from_user(app.ARGS.rpe_pair))[0]
                pa_bids_path = pa_fpath + '.json'
                if not os.path.exists(pa_bids_path):
                    raise MRtrixError('for variable TE data the RPE \
                        image must be accompanies by a bids .json or use\
                        must use the -rpe_te argument to specify the PA echo time')

                pa_bids = json.load(open(pa_bids_path))
                te_pa = float(pa_bids['echo_time'])


            id_dwi_match_pa = np.where(te_pa == np.array(echo_times_per_series))[0][0]
            if id_dwi_match_pa.size == 0:
                raise MRtrixError('echo time of reverse phase encoding image does not\
                                  match any of the input echo times, please check.')

            # extract b0s of dwi matching te of PA image
            run.command('mrconvert -coord 3 %s working.mif - | dwiextract -bzero - - | mrmath - mean %s -axis 3' %
                (dwi_metadata['idxlist'][id_dwi_match_pa],
                path.to_scratch('pe_original_meanb0.nii')),
                show=False)
            # extract brain from mean b0
            run.command('bet %s %s -f 0.2 -m' %
                (path.to_scratch('pe_original_meanb0.nii'), 
                path.to_scratch('pe_original_brain')))

            rpe_size = [ int(s) for s in image.Header(path.from_user(app.ARGS.rpe_pair)).size() ]
            if len(rpe_size) == 4:
                run.command('mrconvert -coord 3 0 %s %s' % 
                    (path.from_user(app.ARGS.rpe_pair), path.to_scratch('rpe_b0.nii')))
            else: 
                run.command('mrconvert %s %s' % 
                    (path.from_user(app.ARGS.rpe_pair), path.to_scratch('rpe_b0.nii')))
            run.command('bet %s %s -f 0.20' % 
                (path.to_scratch('rpe_b0.nii'), path.to_scratch('rpe_b0_brain')))


    if app.ARGS.eddy_groups:
        eddy_groups = [int(i) for i in app.ARGS.eddy_groups.rsplit(',')]
        ngroups = max(eddy_groups)
        group_ids = np.arange(1,ngroups+1,1)

        for i in iter(group_ids):
            group_idx = np.where(np.array(eddy_groups) == i)[0]
            volume_idx_list = [dwi_metadata['idxlist'][j] for j in group_idx]
            volume_idx = ','.join(volume_idx_list)

            # get the indices for each series in the group
            group_vol_inds = [list(map(int, j.split(','))) for j in volume_idx_list]
            start = 0
            ginds = []
            for j in range(len(group_vol_inds)):
                gindsi = np.arange(start, start + len(group_vol_inds[j]))
                gindsi_str = ','.join([str(k) for k in gindsi])
                ginds.append(gindsi_str)
                start += len(group_vol_inds[j])            
           
            run.command('mrconvert -coord 3 %s working.mif %s' % 
                (volume_idx,
                path.to_scratch('dwi_pre_eddy_' + str(i) + '.mif')),
                show=False)
            
            if app.ARGS.rpe_pair:
                run.command('dwiextract -bzero %s - | mrmath - mean %s -axis 3' %
                    (path.to_scratch('dwi_pre_eddy_' + str(i) + '.mif'), 
                    path.to_scratch('b0_pre_eddy_' + str(i) + '.nii')))
                run.command('bet %s %s -f 0.2 -m' %
                    (path.to_scratch('b0_pre_eddy_' + str(i) + '.nii'), 
                    path.to_scratch('b0_pre_eddy_' + str(i) + '_brain')))

                run.command('flirt -in %s -ref %s -out %s -dof 6' %
                    (path.to_scratch('rpe_b0_brain'),
                    path.to_scratch('b0_pre_eddy_' + str(i) + '_brain'),
                    path.to_scratch('rpe_to_ref_' + str(i))))

                run.command('flirt -in %s -ref %s -out %s -dof 6' %
                    (path.to_scratch('pe_original_brain' + fsl_suffix), 
                    path.to_scratch('b0_pre_eddy_' + str(i) + '_brain'), 
                    path.to_scratch('pe_to_ref_' + str(i))))

                run.command('mrcat -force -axis 3 %s %s %s' %
                    (path.to_scratch('pe_to_ref_' + str(i) + fsl_suffix),
                    path.to_scratch('rpe_to_ref_' + str(i) + fsl_suffix),
                    path.to_scratch('b0_pair_topup_' + str(i) + '.nii')))

                acqp = np.zeros((2,3))
                if 'i' in pe_dir: acqp[:,0] = 1
                if 'j' in pe_dir: acqp[:,1] = 1
                if 'k' in pe_dir: acqp[:,2] = 1
                if '-' in pe_dir:
                    acqp[0,:] = -acqp[0,:]
                else:
                    acqp[1,:] = -acqp[1,:]
                
                acqp[acqp==-0] = 0
                acqp = np.hstack((acqp, np.array([0.1,0.1])[...,None]))
                np.savetxt(path.to_scratch('topup_acqp.txt'), acqp, fmt="%1.2f")

                # if any of the image dims are odd dont subsample during topup
                odd_dims = [ int(s) for s in image.Header(path.to_scratch('pe_to_ref_' + str(i) + fsl_suffix)).size()[:3] if s % 2 ]
                if np.any(np.array(odd_dims)):
                    flag_no_subsampling = True
                else:
                    flag_no_subsampling = False

                if flag_no_subsampling:
                    run.command('topup --imain=%s --datain=%s --config=b02b0.cnf --subsamp=1 --scale=1 --out=%s --iout=%s' %
                        (path.to_scratch('b0_pair_topup_' + str(i) + '.nii'),
                        path.to_scratch('topup_acqp.txt'),
                        path.to_scratch('topup_results_' + str(i)),
                        path.to_scratch('topup_results_' + str(i) + fsl_suffix)))
                else:
                    run.command('topup --imain=%s --datain=%s --config=b02b0.cnf --scale=1 --out=%s --iout=%s' %
                        (path.to_scratch('b0_pair_topup_' + str(i) + '.nii'),
                        path.to_scratch('topup_acqp.txt'),
                        path.to_scratch('topup_results_' + str(i)),
                        path.to_scratch('topup_results_' + str(i) + fsl_suffix)))
                
                # mask the topup corrected image
                run.command('mrmath %s mean %s -axis 3' %
                            (path.to_scratch('topup_results_' + str(i) + fsl_suffix),
                             path.to_scratch('topup_corrected_' + str(i) + '_mean.nii')
                            ))
                 
                run.command('bet %s %s -f 0.2 -m' %
                    (path.to_scratch('topup_corrected_' + str(i) + '_mean.nii'), 
                    path.to_scratch('topup_corrected_' + str(i) + '_brain')))
                
                run.command('dwifslpreproc -nocleanup -scratch %s -eddy_options %s -rpe_none -eddy_mask %s -topup_files %s -pe_dir %s %s %s' % 
                    (path.to_scratch('eddy_processing_' + str(i)), 
                    eddyopts, 
                    path.to_scratch('topup_corrected_' + str(i) + '_brain_mask' + fsl_suffix),
                    path.to_scratch('topup_results_' + str(i)),
                    pe_dir,
                    path.to_scratch('dwi_pre_eddy_' + str(i) + '.mif'),
                    path.to_scratch('dwi_post_eddy_' + str(i) + '.mif')))
                
            elif app.ARGS.rpe_none:

                run.command('dwifslpreproc -nocleanup -scratch %s -eddy_options %s -pe_dir %s %s %s' % 
                    (path.to_scratch('eddy_processing_' + str(i)), 
                    eddyopts, 
                    pe_dir,
                    path.to_scratch('dwi_pre_eddy_' + str(i) + '.mif'),
                    path.to_scratch('dwi_post_eddy_' + str(i) + '.mif')))
                
            elif app.ARGS.rpe_all:
                ### this needs to be changed for realistic compatiblity to eddy_groups
                run.command('mrconvert -export_grad_mrtrix grad.txt working.mif tmp.mif', show=False)
                run.command('mrconvert -grad grad.txt ' + path.from_user(app.ARGS.rpe_all) + ' dwirpe.mif', show=False)
                run.command('mrcat -axis 3 working.mif dwirpe.mif dwipe_rpe.mif')
                run.command('dwifslpreproc -nocleanup -eddy_options %s -rpe_all -pe_dir %s %s %s' %
                    (eddyopts, 
                     pe_dir, 
                     path.to_scratch('dwi_pre_eddy_' + str(i) + '.mif'),
                     path.to_scratch('dwi_post_eddy_' + str(i) + '.mif')))
                run.function(os.remove,'tmp.mif')

            elif app.ARGS.rpe_header:

                run.command('dwifslpreproc -nocleanup -eddy_options %s -rpe_header %s %s' % 
                (eddyopts,
                path.to_scratch('dwi_pre_eddy_' + str(i) + '.mif'),
                path.to_scratch('dwi_post_eddy_' + str(i) + '.mif')))

            run.command('dwiextract -bzero %s - | mrmath - mean %s -axis 3' %
                    (path.to_scratch('dwi_post_eddy_' + str(i) + '.mif'), 
                    path.to_scratch('b0_post_eddy_' + str(i) + '.nii')))

            # undo grouping:
            for j in range(len(group_idx)):
                run.command('mrconvert -coord 3 %s %s %s' %
                            (ginds[j],
                            path.to_scratch('dwi_post_eddy_' + str(i) + '.mif'),
                            path.to_scratch('dwi_post_eddy_ungrouped_' + str(group_idx[j]) + '.mif')),
                            show=False)
                
        
        registered_data = glob.glob(os.path.join(path.to_scratch(''),'dwi_post_eddy_ungrouped*.mif'))
        registered_data.sort()

        group_alignment(registered_data)
        run.command('mrconvert -force working.mif dwiec.mif', show=False)
        

    # if not variable TE (not using eddy_groups - only run eddy once)
    else:

        if app.ARGS.rpe_pair:

            run.command('dwiextract -bzero dwi.mif - | mrconvert -coord 3 0 - b0pe.nii')
            rpe_size = [ int(s) for s in image.Header(path.from_user(app.ARGS.rpe_pair)).size() ]
            if len(rpe_size) == 4:
                run.command('mrconvert -coord 3 0 ' + path.from_user(app.ARGS.rpe_pair) + ' b0rpe.nii')
            else: 
                run.command('mrconvert ' + path.from_user(app.ARGS.rpe_pair) + ' b0rpe.nii')
            run.command('flirt -in b0rpe.nii -ref b0pe.nii -dof 6 -out b0rpe2pe.nii.gz')
            run.command('mrcat -axis 3 b0pe.nii b0rpe2pe.nii.gz b0_pair_topup.nii')
 
            acqp = np.zeros((2,3))
            if 'i' in pe_dir: acqp[:,0] = 1
            if 'j' in pe_dir: acqp[:,1] = 1
            if 'k' in pe_dir: acqp[:,2] = 1
            if '-' in pe_dir:
                acqp[0,:] = -acqp[0,:]
            else:
                acqp[1,:] = -acqp[1,:]
                
            acqp[acqp==-0] = 0
            acqp = np.hstack((acqp, np.array([0.1,0.1])[...,None]))
            np.savetxt(path.to_scratch('topup_acqp.txt'), acqp, fmt="%1.2f")

            # if any of the image dims are odd dont subsample during topup. might be better off changing this to padding so topup doesnt take forever
            odd_dims = [ int(s) for s in image.Header(path.to_scratch('b0pe.nii')).size()[:3] if s % 2 ]
            if np.any(np.array(odd_dims)):
                flag_no_subsampling = True
            else:
                flag_no_subsampling = False

            if flag_no_subsampling:
                run.command('topup --imain=%s --datain=%s --config=b02b0.cnf --subsamp=1 --scale=1 --out=%s --iout=%s' %
                    (path.to_scratch('b0_pair_topup.nii'),
                    path.to_scratch('topup_acqp.txt'),
                    path.to_scratch('topup_results'),
                    path.to_scratch('topup_results' + fsl_suffix)))
            else:
                run.command('topup --imain=%s --datain=%s --config=b02b0.cnf --scale=1 --out=%s --iout=%s' %
                    (path.to_scratch('b0_pair_topup.nii'),
                    path.to_scratch('topup_acqp.txt'),
                    path.to_scratch('topup_results'),
                    path.to_scratch('topup_results' + fsl_suffix)))
                
            # mask the topup corrected image
            run.command('mrmath %s mean %s -axis 3' %
                        (path.to_scratch('topup_results' + fsl_suffix),
                            path.to_scratch('topup_corrected_mean.nii')
                        ))
                 
            run.command('bet %s %s -f 0.2 -m' %
                (path.to_scratch('topup_corrected_mean.nii'), 
                path.to_scratch('topup_corrected_brain')))
            
            run.command('dwifslpreproc -nocleanup -scratch %s -eddy_options %s -rpe_none -eddy_mask %s -topup_files %s -pe_dir %s %s %s' % 
                    (path.to_scratch('eddy_processing'), 
                    eddyopts, 
                    path.to_scratch('topup_corrected_brain_mask' + fsl_suffix),
                    path.to_scratch('topup_results'),
                    pe_dir,
                    path.to_scratch('working.mif'),
                    path.to_scratch('dwiec.mif')))

        elif app.ARGS.rpe_none:

            run.command('dwiextract -bzero dwi.mif - | mrconvert -coord 3 0 - b0pe.nii')
            run.command('bet %s %s -f 0.2 -m' %
                (path.to_scratch('b0pe.nii'), 
                path.to_scratch('b0_pe_brain')))
            
            run.command('dwifslpreproc -nocleanup -scratch %s -eddy_options %s -rpe_none -eddy_mask %s -pe_dir %s %s %s' % 
                    (path.to_scratch('eddy_processing'), 
                    eddyopts, 
                    path.to_scratch('b0_pe_brain_mask' + fsl_suffix),
                    pe_dir,
                    path.to_scratch('working.mif'),
                    path.to_scratch('dwiec.mif')))
            
        elif app.ARGS.rpe_all:
            # run an initial topup to create a brain mask
            run.command('mrconvert -export_grad_mrtrix grad.txt dwi.mif tmp.mif', show=False)
            run.command('mrconvert -grad grad.txt ' + path.from_user(app.ARGS.rpe_all) + ' dwirpe.mif')

            run.command('dwiextract -bzero working.mif - | mrconvert -coord 3 0 - b0pe.nii')
            run.command('dwiextract -bzero dwirpe.mif - | mrconvert -coord 3 0 - b0rpe.nii')
            run.command('flirt -in b0rpe.nii -ref b0pe.nii -dof 6 -out b0rpe2pe.nii.gz')
            run.command('mrcat -axis 3 b0pe.nii b0rpe2pe.nii.gz b0_pair_topup.nii')
            
            acqp = np.zeros((2,3))
            if 'i' in pe_dir: acqp[:,0] = 1
            if 'j' in pe_dir: acqp[:,1] = 1
            if 'k' in pe_dir: acqp[:,2] = 1
            if '-' in pe_dir:
                acqp[0,:] = -acqp[0,:]
            else:
                acqp[1,:] = -acqp[1,:]
                
            acqp[acqp==-0] = 0
            acqp = np.hstack((acqp, np.array([0.1,0.1])[...,None]))
            np.savetxt(path.to_scratch('topup_acqp.txt'), acqp, fmt="%1.2f")

            # if any of the image dims are odd dont subsample during topup. might be better off changing this to padding so topup doesnt take forever
            odd_dims = [ int(s) for s in image.Header(path.to_scratch('b0pe.nii')).size()[:3] if s % 2 ]
            if np.any(np.array(odd_dims)):
                flag_no_subsampling = True
            else:
                flag_no_subsampling = False

            if flag_no_subsampling:
                run.command('topup --imain=%s --datain=%s --config=b02b0.cnf --subsamp=1 --scale=1 --out=%s --iout=%s' %
                    (path.to_scratch('b0_pair_topup.nii'),
                    path.to_scratch('topup_acqp.txt'),
                    path.to_scratch('topup_results'),
                    path.to_scratch('topup_results' + fsl_suffix)))
            else:
                run.command('topup --imain=%s --datain=%s --config=b02b0.cnf --scale=1 --out=%s --iout=%s' %
                    (path.to_scratch('b0_pair_topup.nii'),
                    path.to_scratch('topup_acqp.txt'),
                    path.to_scratch('topup_results'),
                    path.to_scratch('topup_results' + fsl_suffix)))
                
            # mask the topup corrected image
            run.command('mrmath %s mean %s -axis 3' %
                        (path.to_scratch('topup_results' + fsl_suffix),
                            path.to_scratch('topup_corrected_mean.nii')
                        ))
                 
            run.command('bet %s %s -f 0.2 -m' %
                (path.to_scratch('topup_corrected_mean.nii'), 
                path.to_scratch('topup_corrected_brain')))

            run.command('mrcat -axis 3 working.mif dwirpe.mif dwipe_rpe.mif')
            run.command('dwifslpreproc -nocleanup -eddy_options %s -rpe_all -pe_dir %s -eddy_mask %s dwipe_rpe.mif dwiec.mif' %
                        (eddyopts, 
                        pe_dir,
                        path.to_scratch('topup_corrected_brain_mask' + fsl_suffix)
                        ))
            run.function(os.remove,'tmp.mif')

        elif app.ARGS.rpe_header:

            cmd = ('dwifslpreproc -nocleanup -eddy_options %s -rpe_header working.mif dwiec.mif' % 
                (eddyopts))
            run.command(cmd)

        elif not app.ARGS.rpe_header and not app.ARGS.rpe_all and not app.ARGS.rpe_pair:
            raise MRtrixError("the eddy option must run alongside -rpe_header, -rpe_all, or -rpe_pair option")

    run.command('mrconvert -force dwiec.mif working.mif', show=False)

def run_b1correct(dwi_metadata):
    from mrtrix3 import run

    DWInlist = dwi_metadata['dwi_list']
    idxlist = dwi_metadata['idxlist']

    # b1 bias field correction
    print("...B1 correction...")
    if len(DWInlist) == 1:
        run.command('dwibiascorrect ants -bias biasfield.mif working.mif dwibc.mif')
    else:
        # b1 correction may still need to be done individually for each diffusion series ...
        miflist = []
        for idx,i in enumerate(DWInlist):
            cmd = ('mrconvert -coord 3 %s working.mif dwiprebc%s.mif' % 
                (idxlist[idx], str(idx)))
            run.command(cmd)
            cmd = ('dwibiascorrect ants -bias biasfield%s.mif dwiprebc%s.mif dwibc%s.mif' % 
                (str(idx), str(idx), str(idx)))
            run.command(cmd)
            miflist.append('dwibc' + str(idx) + '.mif')
            DWImif = ' '.join(miflist)
        run.command('mrcat -axis 3 ' + DWImif + ' dwibc.mif')
    run.command('mrconvert -force dwibc.mif working.mif', show=False)

def create_brainmask(fsl_suffix):
    import os, gzip, shutil
    from mrtrix3 import run

    print("...Computing brain mask...")
    run.command('dwiextract -bzero working.mif - | mrmath -axis 3 - mean b0bc.nii')
    # run.command('dwi2mask dwibc.mif - | maskfilter - dilate brain_mask.nii')
    # run.command('fslmaths b0bc.nii -mas brain_mask.nii brain')
    run.command('bet b0bc.nii brain' + fsl_suffix + ' -m -f 0.25')
    if os.path.isfile('brain_mask.nii.gz'):
        with gzip.open('brain_mask' + fsl_suffix, 'rb') as f_in, open('brain_mask.nii', 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
        run.function(os.remove,'brain_mask' + fsl_suffix, show=False)

def run_rice_bias_correct():
    from mrtrix3 import run, app

    print("...Rician Bias correction...")
    if app.ARGS.denoise:
        run.command('mrcalc noisemap.mif -finite noisemap.mif 0 -if lowbnoisemap.mif', show=False)
        run.command('mrcalc working.mif 2 -pow lowbnoisemap.mif 2 -pow -sub -abs -sqrt - | mrcalc - -finite - 0 -if dwirc.mif')
    else:
        run.command('dwidenoise -noise lowbnoisemap.mif -estimator Exp2 dwi.mif dwitmp.mif', show=False)
        app.cleanup('dwitmp.mif')
        run.command('mrcalc working.mif 2 -pow lowbnoisemap.mif 2 -pow -sub -abs -sqrt - | mrcalc - -finite - 0 -if dwirc.mif')
    run.command('mrconvert -force dwirc.mif working.mif')

def run_normalization(dwi_metadata):
    from mrtrix3 import app, run

    DWInlist = dwi_metadata['dwi_list']
    idxlist = dwi_metadata['idxlist']

    print("...B0 Normalization...")
    if app.ARGS.normalize and not app.ARGS.b1correct:
        if not len(DWInlist) == 1:
            miflist = []
            for idx,i in enumerate(DWInlist):
                run.command('mrconvert -coord 3 ' + idxlist[idx] + ' working.mif dwiprenm' + str(idx) + '.mif')
                run.command('dwiextract dwiprenm' + str(idx) + '.mif - -bzero | mrmath - mean mean_bzero_prenm' + str(idx) + '.mif -axis 3')
                run.command('mrfilter -stdev 3 mean_bzero_prenm' + str(idx) + '.mif smooth mean_bzero_sm' + str(idx) + '.mif')
                miflist.append('mean_bzero_sm' + str(idx) + '.mif')
            nmlist = []
            nmlist.append('dwiprenm0.mif')
            for idx,i in enumerate(miflist[1:]):
                run.command('mrcalc ' + miflist[0] + ' ' + i + ' -div ratio0to' + str(idx+1) + '.mif')
                run.command('mrcalc dwiprenm' + str(idx+1) + '.mif ratio0to' + str(idx+1) + '.mif -mult dwinm' + str(idx+1) + '.mif')
                nmlist.append('dwinm' + str(idx+1) + '.mif')
                DWImif = ' '.join(nmlist)
            run.command('mrcat -axis 3 ' + DWImif + ' dwinm.mif')
            run.command('mrconvert -force dwinm.mif working.mif')
    elif app.ARGS.normalize and app.ARGS.b1correct:
        if not len(DWInlist) == 1:
            for idx,i in enumerate(DWInlist):
                run.command('mrconvert -force -coord 3 ' + idxlist[idx] + ' working.mif dwiprenm' + str(idx) + '.mif')
            nmlist = []
            nmlist.append('dwiprenm0.mif')
            for idx,i in enumerate(DWInlist[1:]):
                run.command('mrcalc -force biasfield0.mif biasfield' + str(idx+1) + '.mif' + ' -div ratio0to' + str(idx+1) + '.mif')
                run.command('mrcalc -force dwiprenm' + str(idx+1) + '.mif ratio0to' + str(idx+1) + '.mif -mult dwinm' + str(idx+1) + '.mif')
                nmlist.append('dwinm' + str(idx+1) + '.mif')
                DWImif = ' '.join(nmlist)
            run.command('mrcat -axis 3 ' + DWImif + ' dwinm.mif')
            run.command('mrconvert -force dwinm.mif working.mif')


from re import L
import numpy as np
import os
from scipy.linalg import svd
from numpy_groupies import aggregate
from skimage.util.shape import view_as_windows


def vectorize(image, mask):
    """
    - If the input is 1D or 2D: unpatch it to 3D or 4D using a mask.
    - If the input is 3D or 4D: vectorize to 2D using a mask.
    - Used by all classes so adding as global function
    """
    if mask is None:
        mask = np.ones((image.shape[0], image.shape[1], image.shape[2]), order='F')

    if image.ndim == 1:
        image = np.expand_dims(image, axis=0)

    if image.ndim == 2:
        n = image.shape[0]
        s = np.zeros((mask.shape[0], mask.shape[1], mask.shape[2], n), order='F')
        for i in range(0, n):
            dummy = np.zeros((mask.shape))
            dummy[mask] = image[i,:]
            s[:,:,:,i] = dummy

    if image.ndim == 3:
        image = np.expand_dims(image, axis=-1)

    if image.ndim == 4:
        s = np.zeros((image.shape[-1], np.sum(mask).astype(int)), order='F')
        for i in range(0, image.shape[-1]):
            tmp = image[:,:,:,i]
            s[i,:] = tmp[mask]

    return np.squeeze(s)


def unpad(x, pad_width):
    """
    If there is padding present, remove it.
    """
    slices = []
    for c in pad_width:
        e = None if c[1] == 0 else -c[1]
        slices.append(slice(c[0], e))
    return x[tuple(slices)]


def get_image_dimensions(X):
    """
    M and generally the number of pixels in a patch, N is the number of image volumes
    M and N will vary from case to case, as we generally create a patch with minimum size:
    product(patch_shape) > n measurements
    """
    M = X.shape[0]
    N = X.shape[1]
    
    return (M, N)


def maybe_transpose(M, N, X):
    """
    Compute conjugate matrix and transpose if needed.
    """
    if M < N:
        X = np.conj(X).T
        
    return X


def compute_svd(X):
    """
    Compute singular value decomposition on matrix,
    square result and convert to float32.
    """
    u, vals, v = svd(X, full_matrices=False)
    vals = (vals**2).astype('float32')
    
    return u, v, vals


def select_msv(vals, u, v):
    """
    Select most significant SVD value:
    We are sorting the singular values from largest to smallest, the largest singular 
    value generally corresponds to the largest "thing" in an image that corresponds to 
    signal (often the mean of the image). The smallest SVs will correspond to Gaussian Noise,
    the goal of this function is to automatically detect the cuttoff between noise carrying 
    and signal carryings SVs.
    """
    order = np.argsort(vals)[::-1]
    
    u = u[:,order]
    v = v[:,order]
    vals = vals[order]
    
    return u, v, vals


def init_crop_factor(Mp):
    """
    tn is a crop factor we are not using (cropping out the very smallest SVs) 
    In the case of heteroscedastic noise, There can be a small number of non-noise
    singular values smaller than noise carrying ones. It is sometimes nice to
    therefore include a "crop factor".
    """
    tn = 0
    return (0, np.arange(0,Mp-tn).T)


def create_signal_carrying_vector(Mp):
    """
    Create a vector of potential "signal carrying SVs".
    """
    return np.arange(0,Mp).T


def get_csum(Mp, vals):
    """
    Vector of cumulative sums of singular values.
    """
    return np.cumsum(vals[::-1])[::-1]


def find_signal_carrying_svs(Mp, tn, sigmasq_1, sigmasq_2):
    """
    Find the intersection between cumsum noise variance and MP noise variance.
    If the two curves intersect, let the first intersection correspond to the noisy SV cutoff.
    Otherwise, keep all singular values assuming that none correspond to noise.
    """
    t = np.where(sigmasq_2 < sigmasq_1[:Mp-tn])

    if t[0].any():
        t = t[0][0]
    else:
        t = Mp - 1

    return t


def get_sigma(sigmasq_1, t):
    """
    Sigma is the square root of noise variance at index t.
    """
    return np.sqrt(sigmasq_1[t])


def get_default_kernel(nvols):
    """
    Return a default kernel if not specified in the module configuration. Max size is 9x9x9
    """
    p = np.arange(3, nvols, 2)
    pf = np.where(p**3 >= nvols)[0][0]
    defaultKernel = p[pf]
    return np.array([defaultKernel, defaultKernel, defaultKernel])


def is_diffusion_shelled(grad):
    """
    Determine the number of unique shells in a diffusion series and ensure
    that there is at least one b=0 and one b>0 shell
    """
    b0inds = grad[:,-1] == 0
    if sum(b0inds) > 0:
        at_least_one_bzero = True
    else:
        at_least_one_bzero = False

    bgt0inds = grad[:,-1] > 0
    if sum(bgt0inds) > 0:
        at_least_one_nonbzero = True
    else:
        at_least_one_nonbzero = False

    if at_least_one_bzero and at_least_one_nonbzero:
        return True
    else:
        return False


def compute_noise_variance_vector(vals, Mp, tn, rangeMP):
    """
    Compute vector of noise variance according to Marchneco Pastur theory
    """
    rangeData = vals[:Mp-tn] - vals[Mp-1]
    sigmasq_2 = rangeData / rangeMP
        
    return sigmasq_2


def veraart(Mp, Np, tn, ptn, p, csum):
    """
    Compute veraart
    """
    sigmasq_1 = csum / ((Mp - p) * Np)
    rangeMP = 4 * np.sqrt((Mp - ptn) * (Np - tn))
    
    return (sigmasq_1, rangeMP)


def cordero_grande(Mp, Np, ptn, p, csum):
    """
    Compute Cordero-Grande
    """
    sigmasq_1 = csum / ((Mp - p) * (Np - p))
    rangeMP = 4 * np.sqrt((Mp - ptn) * (Np - ptn))
    
    return (sigmasq_1, rangeMP)


def jespersen(Mp, Np, tn, p, csum):
    """
    Compute Jespersen
    """
    sigmasq_1 = csum / ((Mp - p) * (Np - p))
    rangeMP = 4 * np.sqrt((Np - tn) * (Mp))
    
    return (sigmasq_1, rangeMP)


def eig_shrink(vals, gamma):
    """
    eig_shrink
    """
    t = 1 + np.sqrt(gamma)
    s = np.zeros((vals.shape))
    x = vals[vals > t]
    s[vals > t] = np.sqrt((x**2 - gamma - 1)**2 - 4*gamma) / x
    return np.diag(s)


def crop_image(vals, Mp, Np, u, v, sigma):
    """
    Crop SVs corresponding to noise directly.
    """
    vals_norm = np.sqrt(vals)/(np.sqrt(Mp) * sigma)
    vals_frobnorm = (np.sqrt(Mp)*sigma) * eig_shrink(vals_norm, Np/Mp)
    return np.matrix(u) * vals_frobnorm * np.matrix(v)


def shrink_image(vals, u, v, t):
    """
    Use shrinkage to minimize the influence of noise carrying SVs proportionally.
    """
    vals[t:] = 0
    return np.matrix(u) * np.diag(np.sqrt(vals)) * np.matrix(v)


def compute_weighted_patchav(wp, signal, temp, nvoxels, dims):
    """
    Weighted patch averaging is performed to account for the contribution of
    voxels that occur in multiple patches.
    """
    return aggregate(
        temp, wp*signal, func='sum', size=nvoxels, dtype=signal.dtype
    ).reshape(dims)


def get_weighted_patch_average(temp, dims, nvoxels, kernel, step):
    """
    Get distance based wights for patch averaging.
    """

    (i, j, k) = np.unravel_index(temp, dims)

    distance = (i - np.mean(i, axis=1, keepdims=True)) ** 2 + \
        (j - np.mean(j, axis=1, keepdims=True)) ** 2 + \
        (k - np.mean(k, axis=1, keepdims=True)) ** 2

    count = np.histogram(temp, np.arange(nvoxels + 1))[0]

    denominator = view_as_windows(
        compute_weighted_patchav(1, distance.flatten(), temp.flatten(), nvoxels, dims),
        kernel, step
    ).reshape(-1, np.prod(kernel))
    
    cbl = view_as_windows(
        count.reshape(dims), kernel, step
    ).reshape(-1, np.prod(kernel))

    num = denominator - distance
    wp = num / (denominator * (cbl - 1) + np.finfo(float).eps)
    wp[cbl == 1] = 1
    wp = wp.flatten()

    return wp



"""
Utilities for checking compatibility of inputs to designer

Includes:
splitext_: a function to split .gz compressed images

get_input_info: a function to get datatype, gradient info, and image metadata for each input series

assert_inputs: asserts that any manual inputs match those found in the bids json, if avilable

convert_input_data: Merge input series and gradient info, setting up the working pipeline
"""

def splitext_(path):
    import os
    for ext in ['.tar.gz', '.tar.bz2','.nii.gz']:
        if path.endswith(ext):
            return path[:-len(ext)], path[-len(ext):]
    return os.path.splitext(path)

def get_input_info(input, fslbval, fslbvec, bids):
    import os
    from mrtrix3 import image, MRtrixError

    UserCpath = input.rsplit(',')
    DWIlist = [os.path.realpath(i) for i in UserCpath]
    
    isdicom = False
    for i in DWIlist:
        if not os.path.exists(i):
            print('cannot find input ' + i)
            quit()
        if os.path.isdir(i):
            format = image.Header(i).format()
            if format == 'DICOM':
                isdicom = True
            else:
                raise MRtrixError('input is a directory but does not contain DICOMs, quitting')

    DWIflist = [splitext_(i) for i in DWIlist]
    DWInlist = [i[0] for i in DWIflist]
    DWIext = [i[1] for i in DWIflist]

    if not fslbval:
        bvallist = [i + '.bval' for i in DWInlist]
    else:
        UserBvalpath = fslbval.rsplit(',')
        bvallist = [os.path.realpath(i) for i in UserBvalpath]
    if not fslbvec:
        bveclist = [i + '.bvec' for i in DWInlist]
    else:
        UserBvecpath = fslbvec.rsplit(',')
        bveclist = [os.path.realpath(i) for i in UserBvecpath]

    if not bids:
        bidslist = [i + '.json' for i in DWInlist]
    else:
        UserBidspath = bids.resplit(',')
        bidslist = [os.path.realpath(i) for i in UserBidspath]

    dwi_metadata = {
        'isdicom': isdicom, 
        'dwi_ext': DWIext,
        'bvecs': bveclist,
        'bvals': bvallist,
        'dwi_list': DWInlist,
        'bidslist': bidslist}

    return dwi_metadata

def convert_pe_dir_to_ijk(args_pe_dir):
    from mrtrix3 import MRtrixError

    avail_pe_dirs = [None,'1','-1','2','-2','3','-3','i','i-','j','j-','k','k-','LR','RL','AP','PA','IS','SI']

    if not args_pe_dir in avail_pe_dirs:
        raise MRtrixError('unsupported PE direction - choose one of (1,i,LR,-1,i-, etc...')

    if args_pe_dir == '1':
        pe_dir_app = 'i'
    elif args_pe_dir == '2':
        pe_dir_app = 'j'
    elif args_pe_dir == '3':
        pe_dir_app = 'k'
    elif args_pe_dir == '-1':
        pe_dir_app = 'i-'
    elif args_pe_dir == '-2':
        pe_dir_app = 'j-'
    elif args_pe_dir == '-3':
        pe_dir_app = 'k-'
    elif args_pe_dir == 'RL':
        pe_dir_app = 'i-'
    elif args_pe_dir == 'LR':
        pe_dir_app = 'i'
    elif args_pe_dir == 'AP':
        pe_dir_app = 'j-'
    elif args_pe_dir == 'PA':
        pe_dir_app = 'j'
    elif args_pe_dir == 'IS':
        pe_dir_app = 'k'
    elif args_pe_dir == 'SI':
        pe_dir_app = 'k-'
    elif args_pe_dir == 'i':
        pe_dir_app = 'i'
    elif args_pe_dir == 'j':
        pe_dir_app = 'j'
    elif args_pe_dir == 'k':
        pe_dir_app = 'k'
    elif args_pe_dir == 'i-':
        pe_dir_app = 'i-'
    elif args_pe_dir == 'j-':
        pe_dir_app = 'j-'
    elif args_pe_dir == 'k-':
        pe_dir_app = 'k-'
    else:
        pe_dir_app = None

    return pe_dir_app

def assert_inputs(dwi_metadata, args_pe_dir, args_pf):
    import json
    from mrtrix3 import app, MRtrixError
    import numpy as np

    bidslist = dwi_metadata['bidslist']
    try:
        bids = [json.load(open(i)) for i in bidslist]
    
        TE_bids = []
        pe_dir = []
        pf = []
        for i in bids:
            TE_bids.append(i['EchoTime'])
            pe_dir.append(i['PhaseEncodingDirection'])
            try:
                pf.append(i['PartialFourier'])
            except:
                pf.append(1)

        if not all(x == pf[0] for x in pf):
            raise MRtrixError('input series have different partial fourier factors,', +
                            'series should be processed separately')
        pf_bids = pf[0]

        if not all(x == pe_dir[0] for x in pe_dir):
            raise MRtrixError('input series have different phase encoding directions, series should be processed separately')
        pe_dir_bids = pe_dir[0]
        pe_dir_app = convert_pe_dir_to_ijk(args_pe_dir)

        if app.ARGS.echo_time:
            TE_app = [float(i) for i in app.ARGS.echo_time.rsplit(',')]
        else:
            TE_app = [0] * len(bidslist)
    except:
        print('no bids files identified')
        pe_dir_app = convert_pe_dir_to_ijk(args_pe_dir)
        pe_dir_bids = None
        pf_bids = None
        TE_bids = None
        if app.ARGS.echo_time:
            TE_app = [float(i) for i in app.ARGS.echo_time.rsplit(',')]
        else:
            TE_app = [0] * len(dwi_metadata['dwi_list'])

    if app.ARGS.bshape:
        bshape = [float(i) for i in app.ARGS.bshape.rsplit(',')]
    else:
        bshape = [1] * len(dwi_metadata['dwi_list'])

    if (app.ARGS.echo_time) and (TE_bids) and (TE_bids != TE_app):
        raise MRtrixError('User defined echo times do not match those found in bids .json, please check input data for consistency')
    
    if np.any(np.array(TE_app) > 0):
        TE = TE_app
    elif TE_bids is not None:
        TE = TE_bids
    else:
        TE = TE_app

    # if no partial fourier information is found, assume full sampling
    if not args_pf and not pf_bids:
        args_pf = 1
        pf_bids = 1
    
    if (pe_dir_bids and pe_dir_app) and (pe_dir_app != pe_dir_bids):
        raise MRtrixError('input phase encoding direction and phase encoding direction from bids file do not match')
    elif pe_dir_bids:
        pe_dir = pe_dir_bids
    else:
        pe_dir = pe_dir_app

    if (pf_bids and args_pf) and (float(args_pf) != pf_bids):
        raise MRtrixError('input partial fourier factor and bids PF factor do not match')
    elif pf_bids:
        pf = pf_bids
    else:
        pf = args_pf
    
    def convert_to_float(frac_str):
        try:
            return float(frac_str)
        except ValueError:
            num, denom = frac_str.split('/')
            try:
                leading, num = num.split(' ')
                whole = float(leading)
            except ValueError:
                whole = 0
            frac = float(num) / float(denom)
            return whole - frac if whole < 0 else whole + frac

    dwi_metadata['pe_dir'] = pe_dir
    dwi_metadata['pf'] = convert_to_float(pf)
    dwi_metadata['TE'] = TE
    dwi_metadata['bshape'] = bshape

def convert_input_data(dwi_metadata):
    """
    convert input data to .mif mrtrix format
    concatenates all inputs along 4th dimension
    
    """
    from mrtrix3 import run, image, MRtrixError, app
    import numpy as np
    import os

    miflist = []
    idxlist = []
    telist = []
    bshapelist = []
    dwi_ind_size = [[0,0,0,0]]

    dwi_n_list = dwi_metadata['dwi_list']
    isdicom = dwi_metadata['isdicom']
    bveclist = dwi_metadata['bvecs']
    bvallist = dwi_metadata['bvals']
    dwi_ext = dwi_metadata['dwi_ext']
    te_per_series = dwi_metadata['TE']
    bshape_per_series = dwi_metadata['bshape']

    if len(dwi_n_list) == 1:
        if not isdicom:
            if os.path.exists(bvallist[0]):
                cmd = ('mrconvert -fslgrad %s %s %s%s %s/dwi.mif' % 
                    (bveclist[0], bvallist[0], ''.join(dwi_n_list), ''.join(dwi_ext), app.SCRATCH_DIR))
                run.command(cmd)
            elif ~os.path.exists(bvallist[0]) and ('.mif' in ''.join(dwi_ext)):
                cmd = ('mrconvert %s%s %s/dwi.mif' % 
                    (''.join(dwi_n_list), ''.join(dwi_ext), app.SCRATCH_DIR))
                run.command(cmd)
            else:
                raise MRtrixError('please make sure that inputs are either .nii files accompanied by corresponding .bval and .bvec files, or a .mif file with embedded gradient information')
            
        else:
            cmd = ('mrconvert %s %s/dwi.mif' %
            (''.join(dwi_n_list), app.SCRATCH_DIR))
            run.command(cmd)
        dwi_header = image.Header('%s/dwi.mif' % (app.SCRATCH_DIR))
        dwi_ind_size.append([ int(s) for s in dwi_header.size() ])

    else:
        for idx,i in enumerate(dwi_n_list):
            if not isdicom:
                if os.path.exists(bvallist[idx]):
                    cmd = ('mrconvert -fslgrad %s %s %s%s %s/dwi%s.mif' % 
                        (bveclist[idx], bvallist[idx], i, dwi_ext[idx], app.SCRATCH_DIR, str(idx)))
                    run.command(cmd)
                elif ~os.path.exists(bvallist[idx]) and ('.mif' in dwi_ext[idx]):
                    cmd = ('mrconvert %s%s %s/dwi%s.mif' % 
                        (i, dwi_ext[idx], app.SCRATCH_DIR, str(idx)))
                    run.command(cmd)
                else:
                    raise MRtrixError('please make sure that inputs are either .nii files accompanied by corresponding .bval and .bvec files, or a .mif file with embedded gradient information')
            else:
                cmd = ('mrconvert %s %s/dwi%s.mif' %
                (i, app.SCRATCH_DIR, str(idx)))
                run.command(cmd)
            dwi_header = image.Header('%s/dwi%s.mif' % (app.SCRATCH_DIR, str(idx)))
            dwi_ind_size.append([ int(s) for s in dwi_header.size() ])
            miflist.append('%s/dwi%s.mif' % (app.SCRATCH_DIR, str(idx)))

        DWImif = ' '.join(miflist)
        cmd = ('mrcat -axis 3 %s %s/dwi.mif' % (DWImif, app.SCRATCH_DIR))
        run.command(cmd)

    # get diffusion header info - check to make sure all values are valid for processing
    dwi_header = image.Header('%s/dwi.mif' % (app.SCRATCH_DIR))
    dwi_size = [ int(s) for s in dwi_header.size() ]
    grad = dwi_header.keyval()['dw_scheme']
    grad = [ line for line in grad ]
    grad = [ [ float(f) for f in line ] for line in grad ]
    
    grad = np.array(grad)
    grad[:,-1] = grad[:,-1] / 1000
    
    #stride = dwi_header.strides()
    num_volumes = 1
    if len(dwi_size) == 4:
        num_volumes = dwi_size[3]

    nvols = [i[3] for i in dwi_ind_size]
    for idx,i in enumerate(dwi_n_list):
        if len(dwi_n_list) == 1:
            tmpidxlist = range(0,num_volumes)
        else:
            tmpidxlist = range(sum(nvols[:idx+1]),sum(nvols[:idx+1])+nvols[idx+1])
        idxlist.append(','.join(str(i) for i in tmpidxlist))
        telist.append([te_per_series[idx]] * nvols[idx+1])
        bshapelist.append([bshape_per_series[idx]] * nvols[idx+1])

    telist = [item for sublist in telist for item in sublist]
    bshapelist = [item for sublist in bshapelist for item in sublist]

    dwi_metadata['idxlist'] = idxlist
    dwi_metadata['echo_time_per_volume'] = np.array(telist)
    dwi_metadata['bshape_per_volume'] = np.array(bshapelist)
    dwi_metadata['grad'] = grad

    if grad is None:
        raise MRtrixError('No diffusion gradient table found')
    if not len(grad) == num_volumes:
        raise MRtrixError('Number of lines in gradient table (%s) does not match input image (%s volumes); check your input data' % 
        (str(len(grad)), str(num_volumes)))

    run.command('mrconvert %s/dwi.mif %s/working.mif' % (app.SCRATCH_DIR, app.SCRATCH_DIR), show=False)

def create_shell_table(dwi_metadata):
    from lib.smi import SMI

    bvals = dwi_metadata['grad'][:,-1]
    bvecs = dwi_metadata['grad'][:,:-1]

    smi = SMI(bval=bvals, bvec=bvecs)

    bshape = dwi_metadata['bshape_per_volume']
    echo_time = dwi_metadata['echo_time_per_volume']
    smi.set_bshape(bshape)
    smi.set_echotime(echo_time)

    return smi.group_dwi_in_shells_b_beta_te()





def parallel_outlier_smooth(inds, kernel, outlier_locations, dwi_norm, dwi, smoothlevel):
    import numpy as np
    
    
    k = kernel // 2
    x = inds[0]
    y = inds[1]
    z = inds[2]

    akcpatch = True
    while np.all(akcpatch) == True:
        xmin = 0 if x-k-1 < 0 else x-k-1
        xmax = dwi.shape[0] if x+k > dwi.shape[0] else x+k
        ymin = 0 if y-k-1 < 0 else y-k-1
        ymax = dwi.shape[1] if y+k > dwi.shape[1] else y+k
        zmin = 0 if z-k-1 < 0 else z-k-1
        zmax = dwi.shape[2] if z+k > dwi.shape[2] else z+k
    
        psize = (xmax - xmin) * (ymax - ymin) * (zmax - zmin)
        akcpatch = outlier_locations[xmin:xmax, ymin:ymax, zmin:zmax].flatten()
        k += 2

    ref = np.tile(np.reshape(dwi_norm[x,y,z,:],(1,dwi.shape[-1])),(psize,1))
    patch = np.reshape(dwi_norm[xmin:xmax,ymin:ymax,zmin:zmax,:],(psize, dwi.shape[-1]))
    patchorig = np.reshape(dwi[xmin:xmax,ymin:ymax,zmin:zmax,:],(psize, dwi.shape[-1]))
    intensities = np.sqrt(np.sum((patch-ref)**2, axis=1)) / dwi.shape[-1]
        
    min_idx = np.argsort(intensities)
    min_wgs = intensities[min_idx]
    wgs_max = min_wgs[-1]
    min_wgs[akcpatch] = wgs_max
    
    if not smoothlevel:
        goodidx = min_wgs < np.median(min_wgs)
    else:
        goodidx = min_wgs < np.percentile(min_wgs, smoothlevel)

    min_idx = min_idx[goodidx]
    min_wgs = min_wgs[goodidx]
    wgs_max = np.max(min_wgs)
    wgs_inv = wgs_max - min_wgs
    
    wgs_nrm = wgs_inv/np.sum(wgs_inv)
    wval = (patchorig[min_idx,:] * 
            (wgs_nrm[...,None] @ np.ones((1,dwi.shape[-1])))
            ).sum(axis=0)

    return wval

def refit_or_smooth(outlier_locations, dwi, mask=None, smoothlevel=None, n_cores=-3):
    from joblib import Parallel, delayed
    import numpy as np

    if mask is None:
        outinds = np.array(np.where(outlier_locations == 1))
    else:
        outinds = np.array(np.where(mask == 1))


    dwi_norm = abs(dwi) / np.amax(dwi, axis=(0,1,2))
    dwi_new = dwi.copy()
    kernel = 7
    # for i in range(len(outinds[0])):
    #     wval = parallel_outlier_smooth(outinds[:,i], kernel, outlier_locations, dwi_norm, dwi, smoothlevel)

    wval = (Parallel(n_jobs=n_cores, prefer='processes')
            (delayed(parallel_outlier_smooth)(
                outinds[:,i], kernel, outlier_locations, dwi_norm, dwi, smoothlevel
            ) for i in range(len(outinds[0]))))

    dwi_new[outinds[0,:],outinds[1,:],outinds[2,:],:] = np.array(wval)

    return dwi_new

def save_params(paramDict, nii, model, outdir):
    from ants import from_numpy, image_write
    import os

    params = paramDict.keys()
    for key in params:
        outpath = os.path.join(outdir, ('%s_%s.nii' % (key, model)))
        vol = paramDict[key]
        ndims = vol.ndim

        out = from_numpy(
        paramDict[key], origin=nii.origin[:ndims], spacing=nii.spacing[:ndims], direction=nii.direction[:ndims,:])
        image_write(out, outpath)




#!/usr/bin/env python3
__version__ = "2.0.0"

from lib.designer_input_utils import *
from lib.designer_func_wrappers import *

def usage(cmdline): #pylint: disable=unused-variable
    from mrtrix3 import app #pylint: disable=no-name-in-module, import-outside-toplevel
    cmdline.set_author('Benjamin Ades-Aron (benjamin.ades-aron@nyulangone.org)')
    cmdline.set_synopsis("""
                        Designer by default, with no optional arguments used "designer <input> <output>" will not perform any preprocessing. Each preprocessing step must be chosen using the appropriate option. For example usage please see the Designer documentation at https://nyu-diffusionmri.github.io/docs/designer/examples/\n
                        
                        1. pre-check: concatenate all dwi series and make sure the all diffusion AP images and PA image have the same matrix dimensions/orientations. Ensure input parameters are reasonable, perform a check on diffusion gradient scheme.\n
 						
                        2. Denoising\n
						
                        3. Gibbs ringing correction\n
 						 						
                        4. EPI distortion and Motion Correction\n
 						
                        5. Normalization\n
 						                        
                        6. Rician bias correction\n
                         						 						
                        """)

    cmdline.add_citation('Veraart, J.; Novikov, D.S.; Christiaens, D.; Ades-aron, B.; Sijbers, J. & Fieremans, E. Denoising of diffusion MRI using random matrix theory. NeuroImage, 2016, 142, 394-406, doi: 10.1016/j.neuroimage.2016.08.016',is_external=True)
    cmdline.add_citation('Veraart, J.; Fieremans, E. & Novikov, D.S. Diffusion MRI noise mapping using random matrix theory. Magn. Res. Med., 2016, 76(5), 1582-1593, doi:10.1002/mrm.26059',is_external=True)
    cmdline.add_citation('Kellner, E., et al., Gibbs-Ringing Artifact Removal Based on Local Subvoxel-Shifts. Magnetic Resonance in Medicine, 2016. 76(5): p. 1574-1581.',is_external=True)
    cmdline.add_citation('Koay, C.G. and P.J. Basser, Analytically exact correction scheme for signal extraction from noisy magnitude MR signals. Journal of Magnetic Resonance, 2006. 179(2): p. 317-322.',is_external=True)
    cmdline.add_citation('Andersson, J. L. & Sotiropoulos, S. N. An integrated approach to correction for off-resonance effects and subject movement in diffusion MR imaging. NeuroImage, 2015, 125, 1063-1078', is_external=True)
    cmdline.add_citation('Smith, S. M.; Jenkinson, M.; Woolrich, M. W.; Beckmann, C. F.; Behrens, T. E.; Johansen-Berg, H.; Bannister, P. R.; De Luca, M.; Drobnjak, I.; Flitney, D. E.; Niazy, R. K.; Saunders, J.; Vickers, J.; Zhang, Y.; De Stefano, N.; Brady, J. M. & Matthews, P. M. Advances in functional and structural MR image analysis and implementation as FSL. NeuroImage, 2004, 23, S208-S219', is_external=True)
    cmdline.add_citation('Skare, S. & Bammer, R. Jacobian weighting of distortion corrected EPI data. Proceedings of the International Society for Magnetic Resonance in Medicine, 2010, 5063', is_external=True)
    cmdline.add_citation('Andersson, J. L.; Skare, S. & Ashburner, J. How to correct susceptibility distortions in spin-echo echo-planar images: application to diffusion tensor imaging. NeuroImage, 2003, 20, 870-888', is_external=True)
    cmdline.add_citation('Zhang, Y.; Brady, M. & Smith, S. Segmentation of brain MR images through a hidden Markov random field model and the expectation-maximization algorithm. IEEE Transactions on Medical Imaging, 2001, 20, 45-57', is_external=True)
    cmdline.add_citation('Smith, S. M.; Jenkinson, M.; Woolrich, M. W.; Beckmann, C. F.; Behrens, T. E.; Johansen-Berg, H.; Bannister P. R.; De Luca, M.; Drobnjak, I.; Flitney, D. E.; Niazy, R. K.; Saunders, J.; Vickers, J.; Zhang, Y.; DeStefano, N.; Brady, J. M. & Matthews, P. M. Advances in functional and structural MR image analysis and implementation as FSL. NeuroImage, 2004,23, S208-S219', is_external=True)
    
    cmdline.add_argument('input',  help='The input DWI series. For multiple input series, separate file names with commas (i.e. dwi1.nii,dwi2.nii,...)')
    cmdline.add_argument('output', help='The output basename')
    
    options = cmdline.add_argument_group('Other options for the DESIGNER script')
    options.add_argument('-degibbs', action='store_true', help='Perform (RPG) Gibbs artifact correction. Must include PF factor with -pf (e.g. 6/8, 7/8) and PF dimension -dim (1, 2 or 3 for i, j  or k respectively).')
    options.add_argument('-rician', action='store_true', help='Perform Rician bias correction')
    options.add_argument('-b1correct', action='store_true', help='Include a bias correction step in dwi preprocessing', default=False)
    options.add_argument('-normalize', action='store_true', help='normalize the dwi volume to median b0 CSF intensity of 1000 (useful for multiple dwi acquisitions)', default=False)
    options.add_argument('-mask', action='store_true',help='compute a brain mask prior to tensor fitting to strip skull and improve efficiency')
    options.add_argument('-echo_time',metavar=('<TE1,TE2,...>'),help='specify the echo time used in the acquisition (comma separated list the same length as number of inputs)')
    options.add_argument('-bshape',metavar=('<beta1,beta2,...>'),help='specify the b-shape used in the acquisition (comma separated list the same length as number of inputs)')
    options.add_argument('-pre_align',action='store_true',help='rigidly align each input series to correct for large motion between series')
    options.add_argument('-ants_motion_correction',action='store_true',help='perform rigid motion correction using ANTs (useful for cases where eddy breaks down)')
    options.add_argument('-pe_dir', metavar=('<phase encoding direction>'), help='Specify the phase encoding direction of the input series (required if using the eddy option). Can be a signed axis number (e.g. -0, 1, +2), an axis designator (e.g. RL, PA, IS), or NIfTI axis codes (e.g. i-, j, k)')
    options.add_argument('-pf', metavar=('<PF factor>'), help='Specify the partial fourier factor (e.g. 7/8, 6/8)')
    options.add_argument('-datatype', metavar=('<spec>'), help='If using the "-processing_only" option, you can specify the output datatype. Valid options are float32, float32le, float32be, float64, float64le, float64be, int64, uint64, int64le, uint64le, int64be, uint64be, int32, uint32, int32le, uint32le, int32be, uint32be, int16, uint16, int16le, uint16le, int16be, uint16be, cfloat32, cfloat32le, cfloat32be, cfloat64, cfloat64le, cfloat64be, int8, uint8, bit')
    options.add_argument('-fslbvec',metavar=('<bvecs>'),help='specify bvec path if path is different from the path to the dwi or the file has an unusual extension')
    options.add_argument('-fslbval',metavar=('<bvals>'),help='specify bval path if path is different from the path to the dwi or the file has an unusual extension')
    options.add_argument('-bids',metavar=('<bids1,bids2,...>'),help='specify bids.json path if path is different from the path to the dwi or the file has an unusual extension')
    options.add_argument('-n_cores',metavar=('<ncores>'),help='specify the number of cores to use in parallel tasks, by default designer will use available cores - 2', default=-3)

    mp_options = cmdline.add_argument_group('Options for specifying denoising parameters')
    mp_options.add_argument('-patch2self', action='store_true', help='Perform patch2self')
    mp_options.add_argument('-denoise', action='store_true', help='Perform MPPCA')
    mp_options.add_argument('-shrinkage',metavar=('<shrink>'),help='specify shrinkage type for MPPCA. Options are "threshold" or "(frob)"', default='frob')
    mp_options.add_argument('-algorithm',metavar=('<alg>'),help='specify MP algorithm. Options are "veraart","(cordero-grande)","jespersen"',default='cordero-grande')
    mp_options.add_argument('-extent', metavar=('<size>'), help='MPPCA Denoising extent. Default is 5,5,5')
    mp_options.add_argument('-phase', metavar=('<image>'), help='Diffusion phases - for performing denoising on complex data. This option should not be used alongside "-rician" since denoising complex data reduces the noise floor.', default=None)
    mp_options.add_argument('-adaptive_patch', action='store_true', help='Run MPPCA with adaptive patching')
    
    rpe_options = cmdline.add_argument_group('Options for eddy and to specify the acquisition phase-encoding design')
    rpe_options.add_argument('-eddy', action='store_true', help='run fsl eddy (note that if you choose this command you must also choose a phase encoding option')
    rpe_options.add_argument('-eddy_groups',metavar=('<index1,index2,...'),help='specify how input series should be grouped when running eddy, for use with variable TE or b-shape data. (Comma separated list of integers beginning with 1, i.e. 1,1,1,2,2,2 for 6 series where the first 3 series and second 3 series have different echo times).', default=None)
    rpe_options.add_argument('-rpe_none', action='store_true', help='Specify that no reversed phase-encoding image data is being provided; eddy will perform eddy current and motion correction only')
    rpe_options.add_argument('-rpe_pair', metavar=('<reverse PE b=0 image>'), help='Specify the reverse phase encoding image')
    rpe_options.add_argument('-rpe_all', metavar=('<reverse PE dwi volume>'), help='Specify that ALL DWIs have been acquired with opposing phase-encoding; this information will be used to perform a recombination of image volumes (each pair of volumes with the same b-vector but different phase encoding directions will be combined together into a single volume). The argument to this option is the set of volumes with reverse phase encoding but the same b-vectors as the input image')
    rpe_options.add_argument('-rpe_header', action='store_true', help='Specify that the phase-encoding information can be found in the image header(s), and that this is the information that the script should use')
    rpe_options.add_argument('-rpe_te', metavar=('<echo time (s)>'), help='Specify the echo time of the reverse phase encoded image, if it is not accompanied by a bids .json sidecar.')

def execute(): #pylint: disable=unused-variable
    from mrtrix3 import app, fsl, run, path #pylint: disable=no-name-in-module, import-outside-toplevel
    import pandas as pd
    import os

    # create a temporary directory to store processing files
    app.make_scratch_dir()
    
    # grab the fsl suffix stored in $FSLOUTPUTTYPLE
    fsl_suffix = fsl.suffix()

    # create dict containing metadata from either user input args or bids .json sidecars
    dwi_metadata = get_input_info(
        app.ARGS.input, app.ARGS.fslbval, app.ARGS.fslbvec, app.ARGS.bids)
    
    # ensure inputs are reasonable for subsequent dwi processing
    assert_inputs(dwi_metadata, app.ARGS.pe_dir, app.ARGS.pf)

    # convert input data to .mif format and concatenate
    convert_input_data(dwi_metadata)

    # get a table of b-shells, echo times, and b-shapes
    shell_table = create_shell_table(dwi_metadata)
    shell_rows = ['b-value', 'b-shape', 'n volumes', 'echo time']
    shell_df = pd.DataFrame(data = shell_table, 
                  index = shell_rows)
    print('input DWI data has properties:')
    print(shell_df)

    app.goto_scratch_dir()
    
    # begin pipeline

    # denoising
    if app.ARGS.denoise:
        if app.ARGS.phase:
            phasepath = path.from_user(app.ARGS.phase)
        else:
            phasepath = None

        # by default perform mppca denoising with optional phase denoising for complex data
        run_mppca(
            app.ARGS.extent, phasepath, app.ARGS.shrinkage, app.ARGS.algorithm)
    
    # patch2self can be run either in addition or alternatively to mppca
    if app.ARGS.patch2self:
        run_patch2self()

    # rpg gibbs artifact correction
    if app.ARGS.degibbs:
        run_degibbs(dwi_metadata['pf'], dwi_metadata['pe_dir'])

    # rigid alignment of b0s from separate input series
    if app.ARGS.pre_align:
        run_pre_align(dwi_metadata)

    # rigid alignment of each dwi volume to the prior volume
    if app.ARGS.ants_motion_correction:
        run_ants_moco(dwi_metadata)

    # eddy current, succeptibility, motion correction
    if app.ARGS.eddy:
        run_eddy(shell_table, dwi_metadata)

    # b1 bias correction
    if app.ARGS.b1correct:
        run_b1correct(dwi_metadata)

    # generate a final brainmask
    if app.ARGS.mask or app.ARGS.normalize:
        create_brainmask(fsl_suffix)    

    # rician bias correction
    if app.ARGS.phase and app.ARGS.rician:
        print('phase included, skipping approximated bias correction')
    elif app.ARGS.rician and not app.ARGS.phase:
        run_rice_bias_correct()

    # normalize separate input series (voxelwise) such that b0 images are properly scaled
    if app.ARGS.normalize:
        run_normalization(dwi_metadata)

    run.command('mrinfo -export_grad_fsl dwi_designer.bvec dwi_designer.bval working.mif', show=False)
    run.command('mrconvert -force -datatype float32le working.mif dwi_designer.nii', show=False)

    outpath = path.from_user(app.ARGS.output, True)
    (head, fname) = os.path.split(outpath)
    oname = fname.split(os.extsep)[0]

    if app.ARGS.datatype:
        run.command('mrconvert -force -datatype %s -export_grad_fsl %s %s %s %s' % 
            (app.ARGS.datatype, 
            os.path.join(head, oname + '.bvec'),
            os.path.join(head, oname + '.bval'),
            path.to_scratch('working.mif', True),
            outpath))
    else:
        run.command('mrconvert -force -export_grad_fsl %s %s %s %s' % 
            (
            os.path.join(head, oname + '.bvec'),
            os.path.join(head, oname + '.bval'),
            path.to_scratch('working.mif', True), 
            outpath))

def main():
    import mrtrix3
    mrtrix3.execute() #pylint: disable=no-member






from joblib import Parallel, delayed
import numpy as np
from tqdm import tqdm
from lib.mpunits import vectorize
import os

import scipy.io as sio
import cvxpy as cp

import multiprocessing
import warnings
warnings.filterwarnings("ignore")

class TensorFitting(object):

    def __init__(self, grad, n_cores):
        self.grad = grad
        self.n_cores = n_cores

    def create_tensor_order(self, order):
        """
        Function to create a set of indices describing the order of a 6 parameter or 21 parameter diffusion tensor
        for a diffusion tensor, the default order goes as xx, xy, xz, yy, yz, zz, and so on
        """
        if order == 2:
            cnt = np.array([1, 2, 2, 1, 2, 1], dtype=int)
            ind = np.array(([1, 1], [1, 2], [1, 3], [2, 2], [2, 3], [3, 3])) - 1
        if order == 4:
            cnt = np.array([1, 4, 4, 6, 12, 6, 4, 12, 12, 4, 1, 4, 6, 4, 1], dtype=int)
            ind = np.array(([1,1,1,1],[1,1,1,2],[1,1,1,3],[1,1,2,2],[1,1,2,3],[1,1,3,3],\
                [1,2,2,2],[1,2,2,3],[1,2,3,3],[1,3,3,3],[2,2,2,2],[2,2,2,3],[2,2,3,3],[2,3,3,3],[3,3,3,3])) - 1
        return cnt, ind

    def diffusion_coeff(self, dt, dir):
        """
        Compute the apparent diffusion coefficient
        """
        # compute ADC
        dcnt, dind = self.create_tensor_order(2)
        ndir = dir.shape[0]
        bD = np.tile(dcnt,(ndir, 1)) * dir[:,dind[:, 0]] * dir[:,dind[:, 1]]
        adc = bD @ dt
        return adc


    def kurtosis_coeff(self, dt, dir):
        """
        Function to Compute the apparent kurtosis coefficient
        """
        # compute AKC
        wcnt, wind = self.create_tensor_order(4)
        ndir = dir.shape[0]
        adc = self.diffusion_coeff(dt[:6], dir)
        md = np.sum(dt[np.array([0,3,5])], 0)/3
        bW = np.tile(wcnt,(ndir, 1)) * dir[:,wind[:, 0]] * dir[:,wind[:, 1]] * dir[:,wind[:, 2]] * dir[:,wind[:, 3]]
        akc = bW @ dt[6:]
        akc = (akc * np.tile(md**2, (adc.shape[0], 1))) / (adc**2)
        return akc

    def w_kurtosis_coeff(self, dt, dir):
        """
        Function to compute W form of kurtosis coefficient
        """
        # compute AKC
        wcnt, wind = self.create_tensor_order(4)
        T = np.prod(np.reshape(dir[:,wind],(-1,15,4)), axis=2) @ np.diag(wcnt)
        akc = T @ dt[6:]
        return akc

    def fibonacci_sphere(self, samples=1, randomize=True):
        """
        Function to generate N "samples" evenly spaced points on a sphere
        """
        import random
        rnd = 1
        if randomize:
            rnd = random.random() * samples
        points = []
        offset = 2/samples
        increment = np.pi * (3. - np.sqrt(5.))
        for i in range(samples):
            y = ((i * offset) - 1) + (offset / 2)
            r = np.sqrt(1 - pow(y,2))
            phi = ((i + rnd) % samples) * increment
            x = np.cos(phi) * r
            z = np.sin(phi) * r
            points.append([x,y,z])
        return points

    def radial_sampling(self, dir, n):
        """
        Function to output the radial component of a set of evenly distribted direction vectors
        """
        # get the radial component of a set of directions
        dt = 2*np.pi/n
        theta = np.arange(0,2*np.pi-dt,dt)
        dirs = np.vstack((np.cos(theta), np.sin(theta), 0*theta))
        v = np.hstack((-dir[1], dir[0], 0))
        s = np.sqrt(np.sum(v**2))
        c = dir[2]
        V = np.array([[0, -v[2], v[1]],[v[2], 0, -v[0]],[-v[1], v[0], 0]])
        R = np.eye(3) + V + np.matmul(V,V) * (1-c)/(s**2)
        dirs = np.matmul(R, dirs)
        return dirs 

    def wlls(self, shat, dwi, b, C=None):
        """
        Function to perform pseudoinversion on raw diffusion data
        """
        # compute a wlls fit using weights from inital fit shat
        w = np.diag(shat)
        try:
            if C is None:
                dt = np.linalg.pinv(w @ b) @ (w @ np.log(dwi))
            else:
                x = cp.Variable((22,1))
                A = w @ b
                B = w @ np.log(dwi[...,None])
                objective = cp.Minimize(cp.sum_squares(A @ x - B))
                constraints = [C @ x >= 0]
                prob = cp.Problem(objective, constraints)
                result = prob.solve()
                dt = x.value.squeeze()
                # dt,_ = opt.nnls(w @ b, w @ np.log(dwi))
                # res = opt.lsq_linear(w @ b, w @ np.log(dwi), (-5, 1), method='trf', tol=1e-12, max_iter=220000)
                # dt = res.x
        except:
            dt = np.zeros((22))

        return dt  

    def dti_tensor_params(self, nn):
        """
        Function to compute orthogonal projections of diffusion from a spherical tensor representation
        """
        # compute dti tensor eigenvalues and eigenvectors and sort them
        values, vectors = np.linalg.eig(nn)
        idx = np.argsort(-values)
        values = -np.sort(-values)
        vectors = vectors[:, idx]
        return values, vectors  

    def dki_tensor_params(self, v1, dt, fit_w=False):
        """
        Function to compute orthogonal projections of kurtosis from a spherical tensor representation
        """
        # kurtosis tensor parameters use average directional
        # statistics to approximate ak and rk
        dirs = np.vstack((v1, -v1))
        if fit_w:
            akc = self.w_kurtosis_coeff(dt, dirs)
            ak = np.mean(akc)
            dirs = self.radial_sampling(v1, 256).T
            akc = self.w_kurtosis_coeff(dt, dirs)
            rk = np.mean(akc)
        else:
            akc = self.kurtosis_coeff(dt, dirs)
            ak = np.mean(akc)
            dirs = self.radial_sampling(v1, 256).T
            akc = self.kurtosis_coeff(dt, dirs)
            rk = np.mean(akc)
        return ak, rk    

    def extract_parameters(self, dt, b, mask, extract_dti, extract_dki, fit_w=False):
        """
        Computes parameters MD, AD, RD, FA and Trace of the diffusion tensor, along with color encoded FA
        Computes parameters MK AK and RK of the kurtosis tensor
        """
        # extract all tensor parameters from dt
        # num_cores = multiprocessing.cpu_count()

        DT = np.reshape(np.concatenate(
            (dt[0,:], dt[1,:], dt[2,:], dt[1,:], dt[3,:], dt[4,:], dt[2,:], dt[4,:], dt[5,:])
            ), (3, 3, dt.shape[1]))
        
        # get the trace
        rdwi = np.exp(np.matmul(b[:,1:], dt))
        B = np.round(self.grad[:,-1])
        uB = np.unique(B)
        trace = np.zeros((dt.shape[1], uB.shape[0]))
        for ib in range(0, uB.shape[0]): 
            t = np.where(B == uB[ib])[0]
            trace[:,ib] = np.mean(rdwi[t,:], axis=0)

        nvox = dt.shape[1]
        inputs = range(0, nvox)
        values, vectors = zip(*Parallel(n_jobs=self.n_cores, prefer='processes')\
            (delayed(self.dti_tensor_params)(DT[:,:,i]) for i in inputs))        
        values = np.reshape(np.abs(values), (nvox, 3))
        vectors = np.reshape(vectors, (nvox, 3, 3))

        l1 = vectorize(values[:,0], mask)
        l2 = vectorize(values[:,1], mask)
        l3 = vectorize(values[:,2], mask)
        v1 = vectorize(vectors[:,:,0].T, mask)

        md = (l1+l2+l3)/3
        rd = (l2+l3)/2
        ad = l1
        fa = np.sqrt(1/2)*np.sqrt((l1-l2)**2+(l2-l3)**2+(l3-l1)**2)/np.sqrt(l1**2+l2**2+l3**2)
        trace = vectorize(trace.T, mask)
        fe = np.abs(np.stack((fa*v1[:,:,:,0], fa*v1[:,:,:,1], fa*v1[:,:,:,2]), axis=3))
        
        parameters = {}
        if extract_dti:
            parameters['md'] = md
            parameters['rd'] = rd
            parameters['ad'] = ad
            parameters['fa'] = fa
            parameters['fe'] = fe
            parameters['trace'] = trace

        if extract_dki:
            dirs = np.array(self.fibonacci_sphere(256, True))
            
            if fit_w:
                akc = self.w_kurtosis_coeff(dt, dirs)
                mk = np.mean(akc, axis=0)
                ak, rk = zip(*Parallel(n_jobs=self.n_cores, prefer='processes')\
                    (delayed(self.dki_tensor_params)(vectors[i,:,0], dt[:,i], fit_w=True) for i in inputs))
                ak = np.reshape(ak, (nvox))
                rk = np.reshape(rk, (nvox))
                ak = vectorize(ak, mask) * md**2 / ad**2
                rk = vectorize(rk, mask) * md**2 / rd**2
            else:
                akc = self.kurtosis_coeff(dt, dirs)
                mk = np.mean(akc, axis=0)
                ak, rk = zip(*Parallel(n_jobs=self.n_cores, prefer='processes')\
                    (delayed(self.dki_tensor_params)(vectors[i,:,0], dt[:,i], fit_w=False) for i in inputs))
                ak = np.reshape(ak, (nvox))
                rk = np.reshape(rk, (nvox))
                ak = vectorize(ak, mask)
                rk = vectorize(rk, mask)
            mk = vectorize(mk, mask)

            parameters['mk'] = mk
            parameters['ak'] = ak
            parameters['rk'] = rk
        
        return parameters

    def dki_fit(self, dwi, mask, constraints=None):
        """ 
        Diffusion parameter estimation using weightel linear least squares fitting
        Outputs the 6 parameter diffusion tensor and 21 parameter kurtosis tensor in dt[diffusion,kurtosis]
        Outputs S0 the true quantitative mean signal at zero gradient strength
        the gradient tensor b
        """
        # run the fit
        order = np.floor(np.log(np.abs(np.max(self.grad[:,-1])+1))/np.log(10))
        if order >= 2:
            self.grad[:, -1] = self.grad[:, -1]/1000

        dwi.astype(np.double)
        dwi[dwi <= 0] = np.finfo(np.double).eps

        self.grad = self.grad.astype(np.double)
        normgrad = np.sqrt(np.sum(self.grad[:,:3]**2, 1))
        normgrad[normgrad == 0] = 1

        self.grad[:,:3] = self.grad[:,:3]/np.tile(normgrad, (3,1)).T
        self.grad[np.isnan(self.grad)] = 0

        dcnt, dind = self.create_tensor_order(2)
        wcnt, wind = self.create_tensor_order(4)

        ndwis = dwi.shape[-1]
        bs = np.ones((ndwis, 1))
        bD = np.tile(dcnt,(ndwis, 1))*self.grad[:,dind[:, 0]]*self.grad[:,dind[:, 1]]
        bW = np.tile(wcnt,(ndwis, 1))*self.grad[:,wind[:, 0]]*self.grad[:,wind[:, 1]]*self.grad[:,wind[:, 2]]*self.grad[:,wind[:, 3]]
        b = np.concatenate((bs, (np.tile(-self.grad[:,-1], (6,1)).T * bD), np.squeeze(1/6 * np.tile(self.grad[:,-1], (15,1)).T**2) * bW), 1)

        dwi_ = vectorize(dwi, mask)
        init = np.matmul(np.linalg.pinv(b), np.log(dwi_))
        shat = np.exp(np.matmul(b, init))

        if np.any(constraints):
            dwd = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            mat = sio.loadmat(os.path.join(dwd,'constant','dirs_constrained.mat'))
            dir = mat['dir']

            ndir = dir.shape[0]
            C = []
            if constraints[0] > 0:
                c0 = np.hstack((np.zeros((ndir,1)), 
                    np.tile(dcnt, (ndir, 1)) * dir[:, dind[:,0]] * dir[:,dind[:,1]], 
                    np.zeros((ndir, 15))))
                C.append(c0)
            if constraints[1] > 0:
                c1 = np.hstack((np.zeros((ndir,7)), 
                    np.tile(wcnt, (ndir, 1)) * dir[:, wind[:,0]] * dir[:,wind[:,1]] * dir[:,wind[:,2]] * dir[:,wind[:,3]]))
                C.append(c1)
            if constraints[2] > 0:
                c2 = np.hstack((np.zeros((ndir,1)), 
                    3/np.max(self.grad[:,3]) * np.tile(dcnt, (ndir, 1)) * dir[:, dind[:,0]] * dir[:,dind[:,1]], 
                    -np.tile(wcnt, (ndir, 1)) * dir[:,wind[:,0]] * dir[:,wind[:,1]] * dir[:,wind[:,2]] * dir[:,wind[:,3]]))
                C.append(c2)
            C = np.reshape(C, (ndir*np.sum(constraints), -1))
        else:
            C = None

        inputs = tqdm(range(0, dwi_.shape[1]))
        # num_cores = multiprocessing.cpu_count()

        # for i in inputs:
        #     self.wlls(shat[:,i], dwi_[:,i], b, C)
         
        dt = Parallel(n_jobs=self.n_cores,prefer='processes')\
            (delayed(self.wlls)(shat[:,i], dwi_[:,i], b, C) for i in inputs)
        dt = np.reshape(dt, (dwi_.shape[1], b.shape[1])).T

        s0 = np.exp(dt[0,:])
        dt = dt[1:,:]
        D_apprSq = 1/(np.sum(dt[(0,3,5),:], axis=0)/3)**2
        dt[6:,:] = dt[6:,:]*np.tile(D_apprSq, (15,1))
        return dt, s0, b

    def dti_fit(self, dwi, mask):

        # run the fit
        order = np.floor(np.log(np.abs(np.max(self.grad[:,-1])+1))/np.log(10))
        if order >= 2:
            self.grad[:, -1] = self.grad[:, -1]/1000

        dwi.astype(np.double)
        dwi[dwi <= 0] = np.finfo(np.double).eps

        self.grad = self.grad.astype(np.double)
        normgrad = np.sqrt(np.sum(self.grad[:,:3]**2, 1))
        normgrad[normgrad == 0] = 1

        self.grad[:,:3] = self.grad[:,:3]/np.tile(normgrad, (3,1)).T
        self.grad[np.isnan(self.grad)] = 0

        dcnt, dind = self.create_tensor_order(2)

        ndwis = dwi.shape[-1]
        bs = np.ones((ndwis, 1))
        bD = np.tile(dcnt,(ndwis, 1))*self.grad[:,dind[:, 0]]*self.grad[:,dind[:, 1]]
        b = np.concatenate((bs, (np.tile(-self.grad[:,-1], (6,1)).T * bD)),1)

        dwi_ = vectorize(dwi, mask)
        init = np.matmul(np.linalg.pinv(b), np.log(dwi_))
        shat = np.exp(np.matmul(b, init))

        inputs = tqdm(range(0, dwi_.shape[1]))
        # num_cores = multiprocessing.cpu_count()
        
        dt = Parallel(n_jobs=self.n_cores, prefer='processes')\
            (delayed(self.wlls)(shat[:,i], dwi_[:,i], b) for i in inputs)
        dt = np.reshape(dt, (dwi_.shape[1], b.shape[1])).T

        s0 = np.exp(dt[0,:])
        dt = dt[1:,:]
        return dt, s0, b

    def compute_outliers(self, dt, dir):
        akc = self.kurtosis_coeff(dt, dir)
        return np.any((akc < -1) | (akc > 10), axis=0)

    def outlierdetection(self, dt, mask, dir):

        nvxls = dt.shape[1]
        akc_mask = np.zeros((nvxls))
        nblocks = 200

        try:
            N = 10000
            akc_mask = Parallel(n_jobs=self.n_cores, prefer='threads')\
                (delayed(self.compute_outliers)
                (dt, dir[int(N/nblocks*(i-1)):int(N/nblocks*i),:]) for i in range(1, nblocks + 1)
                )
        except:
            N = 1000
            akc_mask = Parallel(n_jobs=8)\
                (delayed(self.compute_outliers)
                (dt, dir[int(N/nblocks*(i-1)):int(N/nblocks*i),:]) for i in range(1, nblocks + 1)
                )

        return np.sum(akc_mask, axis=0)





"""A Python class containing an implimentation of MPPCA denoising.
        
    Inputs are a 4D image with dimentions (X x Y x Z x N)
    
    Usage:
    import mpdenoise as mp
    imgdn, sigma, nparameters = mp.denoise(img, kernel=[7,7,7], M=60, shrinkage='frobenius')
    
    LICENCE
    Authors: Benjamin Ades-Aron (Benjamin.Ades-Aron@nyulangone.org)
    Copyright (c) 2016 New York University
    
    Permission is hereby granted, free of charge, to any non-commercial entity
    ('Recipient') obtaining a copy of this software and associated
    documentation files (the 'Software'), to the Software solely for
    non-commercial research, including the rights to use, copy and modify the
    Software, subject to the following conditions:
    
    1. The above copyright notice and this permission notice shall be
    included by Recipient in all copies or substantial portions of the
    Software.
    
    2. THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIESOF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
    NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BELIABLE FOR ANY CLAIM,
    DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
    OTHERWISE, ARISING FROM, OUT OF ORIN CONNECTION WITH THE SOFTWARE OR THE
    USE OR OTHER DEALINGS IN THE SOFTWARE.
    
    3. In no event shall NYU be liable for direct, indirect, special,
    incidental or consequential damages in connection with the Software.
    Recipient will defend, indemnify and hold NYU harmless from any claims or
    liability resulting from the use of the Software by recipient.

    4. Neither anything contained herein nor the delivery of the Software to
    recipient shall be deemed to grant the Recipient any right or licenses
    under any patents or patent application owned by NYU.

    5. The Software may only be used for non-commercial research and may not
    be used for clinical care.

    6. Any publication by Recipient of research involving the Software shall
    cite the references listed below.

    REFERENCES
    Veraart, J.; Fieremans, E. & Novikov, D.S. Diffusion MRI noise mapping
    using random matrix theory Magn. Res. Med., 2016, early view, doi:
    10.1002/mrm.26059
    """

import os, sys
import numpy as np
import multiprocessing
from joblib import Parallel, delayed
from tqdm import tqdm
import bottleneck as bn
import scipy.linalg
#import jax.scipy as jsp
#from jax import jit
import warnings

warnings.filterwarnings('ignore')

class MP(object):
    def __init__(self, dwi, kernel, patchtype, patchsize, shrinkage, algorithm):
        if dwi.ndim > 4:
            self.coil = True
        else:
            self.coil = False

        if shrinkage is None:
            self.shrink = 'threshold'
        else:
            self.shrink = shrinkage

        if algorithm is None:
            self.algo = 'jespersen'
        else:
            self.algo = algorithm

        if kernel is None:
            nvols = dwi.shape[-1]
            p = np.arange(3, nvols, 2)
            pf = np.where(p**3 >= nvols)[0][0]
            defaultKernel = p[pf]
            kernel = np.array([defaultKernel,defaultKernel,defaultKernel])
        else:
            kernel = np.array(kernel)

        kernel = kernel+np.mod(kernel,2)-1
        self.kernel = kernel.astype(int)
        k = self.kernel // 2
        
        if patchtype is None:
            self.patchtype = 'box'
        else:
            self.patchtype = patchtype

        if patchsize is None:
            self.patchsize = np.prod(kernel)
        else:
            self.patchsize = patchsize

        if self.patchtype == 'box':
            self.pos_distances = None
        elif self.patchtype == 'nonlocal':
            if self.patchsize >= np.prod(kernel):
                print('Warning: selecting sane default nonlocal patch size')
                self.patchsize = np.floor(np.prod(kernel) - 0.2*np.prod(kernel))
            elif self.patchsize <= dwi.shape[-1]:
                print('Warning: selecting sane default nonlocal patch size')
                self.patchsize = dwi.shape[-1] + 1
            if not isinstance(self.patchsize, int):
                self.patchsize = self.patchsize.astype(int)
            pi, pj, pk = np.where(np.ones((self.kernel)))
            patchcoords = np.vstack((pi,pj,pk)).T
            self.pos_distances = np.sum((patchcoords - k)**2, axis=1)
                    
        if self.coil:
            pwidth = (k[0],k[0]), (k[1],k[1]), (k[2],k[2]), (0,0), (0,0)
        else:
            pwidth = (k[0],k[0]), (k[1],k[1]), (k[2],k[2]), (0,0)
        self.pwidth = pwidth
        self.dwi = np.pad(dwi, pad_width=pwidth, mode='wrap')

        print('Denoising data with parameters:')
        print('kernel     = ' + str(self.kernel))
        print('patch type = ' + str(self.patchtype))
        print('patch size = ' + str(self.patchsize))
        print('shrinkage  = ' + str(self.shrink))
        print('algorithm  = ' + str(self.algo))
        print('coil level = ' + str(self.coil))
            
    def box_patch(self, dwi, coords):
        # extracts a patch of size kx x ky x kz from the padded input at specified coords
        k = self.kernel // 2
        X = dwi[coords[0]-k[0]:coords[0]+k[0]+1, coords[1]-k[1]:coords[1]+k[1]+1, coords[2]-k[2]:coords[2]+k[2]+1, ...]
        return X

    def normalize(self, im):
        im = ((im - np.min(im)) * (1/(np.max(im) - np.min(im)) * 1.0))
        return im

    def refine(self, X_tmp, coords):
        refval = X_tmp[np.prod(self.kernel)//2,...]
        if self.coil:
            int_distances = 1/(X_tmp.shape[-1]*X_tmp.shape[-2]) * np.sum((X_tmp - refval)**2, axis=(1,2))
        else:
            int_distances = 1/(X_tmp.shape[-1]) * np.sum((X_tmp - refval)**2, axis=1)

        wdists = (self.pos_distances * int_distances)
        iidx = bn.argpartition(wdists, self.patchsize)[:self.patchsize]
        minind = np.argmin(self.pos_distances[iidx])

        # debug = True
        # if debug:
        #     print('sup')
        #     k = self.kernel // 2
        #     sx, sy, sz, N = self.dwi.shape
        #     mask = np.zeros((sx,sy,sz))
        #     mask[coords[0]-k[0]:coords[0]+k[0]+1, coords[1]-k[1]:coords[1]+k[1]+1, coords[2]-k[2]:coords[2]+k[2]+1] = 1
        #     iidx = bn.argpartition(wdists, self.patchsize)[:self.patchsize]
        #     pos_im = np.reshape(self.pos_distances,self.kernel)
        #     int_im = np.reshape(int_distances,self.kernel)
        #     w_im = np.reshape(wdists,self.kernel)
                
        #     pmask = mask.copy()
        #     pmask[coords[0]-k[0]:coords[0]+k[0]+1, coords[1]-k[1]:coords[1]+k[1]+1, coords[2]-k[2]:coords[2]+k[2]+1] = pos_im
        #     imask = mask.copy()
        #     imask[coords[0]-k[0]:coords[0]+k[0]+1, coords[1]-k[1]:coords[1]+k[1]+1, coords[2]-k[2]:coords[2]+k[2]+1] = int_im
        #     wmask = mask.copy()
        #     wmask[coords[0]-k[0]:coords[0]+k[0]+1, coords[1]-k[1]:coords[1]+k[1]+1, coords[2]-k[2]:coords[2]+k[2]+1] = w_im

        #     minw = np.partition(wdists, self.patchsize)[self.patchsize]
        #     omask = wmask.copy()
        #     omask[wmask>minw] = 1

        #     import matplotlib.pyplot as plt
        #     plt.imshow(np.squeeze(self.dwi[coords[0],:,:,55]))
        #     alphas = np.squeeze(mask[coords[0],:,:])
        #     plt.imshow(alphas, alpha=0.5*alphas)
        #     plt.show()

        #     plt.imshow(np.squeeze(self.dwi[coords[0],:,:,55]))
        #     pd = np.squeeze(pmask[coords[0],:,:])
        #     plt.imshow(1-pd, alpha=0.75*alphas)
        #     plt.show()

        #     plt.imshow(np.squeeze(self.dwi[coords[0],:,:,55]))
        #     id = np.squeeze(imask[coords[0],:,:])
        #     plt.imshow(1-id, alpha=0.75*alphas)
        #     plt.show()

        #     plt.imshow(np.squeeze(self.dwi[coords[0],:,:,55]))
        #     wd = np.squeeze(wmask[coords[0],:,:])
        #     plt.imshow(1-wd, alpha=0.75*alphas)
        #     plt.show()

        #     plt.imshow(np.squeeze(self.dwi[coords[0],:,:,55]))
        #     od = np.squeeze(omask[coords[0],:,:])
        #     plt.imshow(1-od,alpha=alphas*(1-od))
        #     plt.show()

        return iidx, minind

    def unpad(self, x, pad_width):
        slices = []
        for c in pad_width:
            e = None if c[1] == 0 else -c[1]
            slices.append(slice(c[0], e))
        return x[tuple(slices)]

    def padded_sampling(self, mask):
        # outputs a grid x, y, z of which coordinates to loop over when processing
        sx, sy, sz = mask.shape
        k = self.kernel // 2
        mask[:k[0],:,:] = 0
        mask[sx-k[0]:,:,:] = 0
        mask[:,:k[1],:] = 0
        mask[:,sy-k[1]:,:] = 0
        mask[:,:,:k[2]] = 0
        mask[:,:,sz-k[2]:] = 0
        self.mask = mask
        x, y, z = np.where(mask==1)
        return x.astype(int), y.astype(int), z.astype(int)

    def sampling(self, mask):
        # outputs a grid x, y, z of which coordinates to loop over when processing
        x, y, z = np.where(mask==1)
        return x.astype(int), y.astype(int), z.astype(int)

    def eig_shrink(self, vals, gamma):
        t = 1 + np.sqrt(gamma)
        s = np.zeros((vals.shape))
        x = vals[vals > t]
        s[vals > t] = np.sqrt((x**2 - gamma - 1)**2 - 4*gamma) / x
        return np.diag(s)

    # def jsvd(self, X):
    #     u,vals,v = jsp.linalg.svd(X, full_matrices=False, compute_uv=True, check_finite=False, overwrite_a=True)
    #     return u,vals,v
    
    def denoise(self, coords):
        X = self.box_patch(self.dwi, coords)

        if self.coil:
            X = np.reshape(X,(np.prod(self.kernel), self.dwi.shape[-2], self.dwi.shape[-1]))
        else:
            X = np.reshape(X,(np.prod(self.kernel), self.dwi.shape[-1]))

        if self.patchtype == 'nonlocal':
            Xn = self.normalize(X)
            nonlocalinds, minind = self.refine(Xn, coords)
            X = X[nonlocalinds,...]
        else:
            minind = np.prod(self.kernel)//2

        if self.coil:
            X = X.reshape(self.patchsize*self.dwi.shape[-2], self.dwi.shape[-1])
        
        M = X.shape[0]
        N = X.shape[1]
        Mp = np.min((M,N))
        Np = np.max((M,N))
        
        if M < N:  
            X = np.conj(X).T
        
        #try:
        #u,vals,v = self.svdfunc(X)
        u,vals,v = scipy.linalg.svd(X, full_matrices=False)
        vals = (vals**2).astype('float32')

        order = np.argsort(vals)[::-1]
        u = u[:,order]
        v = v[:,order]
        vals = vals[order]

        tn = 0
        ptn = np.arange(0,Mp-tn).T
        p = np.arange(0,Mp).T

        csum = np.cumsum(vals[::-1])[::-1]
        if self.algo == 'veraart':
            sigmasq_1 = csum / ((Mp-p)*Np)
            rangeMP = 4*np.sqrt((Mp-ptn)*(Np-tn))
        elif self.algo == 'cordero-grande':
            sigmasq_1 = csum / ((Mp-p)*(Np-p))
            rangeMP = 4*np.sqrt((Mp-ptn)*(Np-ptn))
        elif self.algo == 'jespersen':
            sigmasq_1 = csum / ((Mp-p)*(Np-p))
            rangeMP = 4*np.sqrt((Np-tn)*(Mp))

        rangeData = vals[:Mp-tn] - vals[Mp-1]
        sigmasq_2 = rangeData / rangeMP

        t = np.where(sigmasq_2 < sigmasq_1[:Mp-tn])

        if t[0].any():
            t = t[0][0]
        else:
            t = Mp-1
        
        sigma = np.sqrt(sigmasq_1[t])
        npars = t
        if self.shrink == 'threshold':
            vals[t:] = 0
            s = np.matrix(u) * np.diag(np.sqrt(vals)) * np.matrix(v)
        else:
            vals_norm = np.sqrt(vals)/(np.sqrt(Mp)*sigma)
            vals_frobnorm = (np.sqrt(Mp)*sigma) * self.eig_shrink(vals_norm, Np/Mp)
            s = np.matrix(u) * vals_frobnorm * np.matrix(v) 
        
        if M<N:
            s = np.conj(s).T

        # except:
        #     #print('Warning: no MP fit')
        #     s = X
        #     sigma = np.nan
        #     npars = np.nan
        #     if M<N:
        #         s = np.conj(s).T 

        if self.coil:
            signal = np.squeeze(s[minind::self.dwi.shape[-2], :])
        else:    
            signal = np.squeeze(s[minind, :]) 

        return signal, sigma, npars

    def process(self):
        sx, sy, sz, N = self.dwi.shape
        mask = np.ones((sx,sy,sz))

        x, y, z = self.padded_sampling(mask)
        xsize = int(x.size)
        coords = np.vstack((x,y,z))
        
        inputs = tqdm(range(0, xsize))
        num_cores = multiprocessing.cpu_count()
        
        # # parallel
        signal, sigma, npars = zip(*Parallel(n_jobs=num_cores, prefer='processes')\
           (delayed(self.denoise)(coords[:,i]) for i in inputs))
        
    #   #  serial
        # k = self.kernel // 2
        # #for t in inputs:
        # crds = np.array([48+k[0], 38+k[1], 25+k[2]])
        # a, b, c = self.denoise(crds)
           
        # reconstruct original data matrix
        Sigma = np.zeros((sx, sy, sz))
        Npars = np.zeros((sx, sy, sz))
        Signal = np.zeros((sx, sy, sz, N))
        for nn in inputs:
            Sigma[x[nn], y[nn], z[nn]] = sigma[nn]
            Npars[x[nn], y[nn], z[nn]] = npars[nn]
            Signal[x[nn], y[nn], z[nn], :] = signal[nn]

        Signal = self.unpad(Signal, self.pwidth)
        Npars = self.unpad(Npars, self.pwidth[:][:-1])
        Sigma = self.unpad(Sigma, self.pwidth[:][:-1])
        return Signal, Sigma, Npars

def denoise(img, kernel=None, patchtype=None, patchsize=None, shrinkage=None, algorithm=None):
    mp = MP(img, kernel, patchtype, patchsize, shrinkage, algorithm)
    Signal, Sigma, Npars = mp.process()
    return Signal, Sigma, Npars

# if __name__ == "__main__":
#     denoise()
    



#!/usr/bin/env python3
import os

from lib.designer_input_utils import get_input_info, convert_input_data, create_shell_table, assert_inputs
from lib.designer_fit_wrappers import refit_or_smooth, save_params

def usage(cmdline): #pylint: disable=unused-variable
    from mrtrix3 import app #pylint: disable=no-name-in-module, import-outside-toplevel
    cmdline.set_author('Benjamin Ades-Aron (benjamin.ades-aron@nyulangone.org)')
    cmdline.set_synopsis("""1. pre-check: concatenate all dwi series and make sure the all diffusion AP images and PA image have the same matrix dimensions/orientations \n 
 						
                        2. Run fitting: Available options include DTI (TMI will look for b0 and b1000 images by default.\n
                            DKI (constrained fitting is available using the -fit_constraints option).\n
                            W-DKI (DKI parameters using the W definition instead of K).\n
                            WMTI (using dipy implimentation from Rafael Henriques).\n
                            SMI (From Santiago Coelho).\n
                        """)

    cmdline.add_citation('Veraart, J.; Novikov, D.S.; Christiaens, D.; Ades-aron, B.; Sijbers, J. & Fieremans, E. Denoising of diffusion MRI using random matrix theory. NeuroImage, 2016, 142, 394-406, doi: 10.1016/j.neuroimage.2016.08.016',is_external=True)

    cmdline.add_argument('input',  help='The input DWI series. For multiple input series, separate file names with commas (i.e. dwi1.nii,dwi2.nii,...)')
    cmdline.add_argument('output', help='The output directory (includes diffusion parameters, kurtosis parameters and processed dwi) unless option -processing_only is used, in which case this is the output basename')
    
    options = cmdline.add_argument_group('Other options for the TMI script')    
    options.add_argument('-mask', metavar=('<path>'),help='path to mask previously computed by designer script')
    options.add_argument('-datatype', metavar=('<spec>'), help='Valid options are float32, float32le, float32be, float64, float64le, float64be, int64, uint64, int64le, uint64le, int64be, uint64be, int32, uint32, int32le, uint32le, int32be, uint32be, int16, uint16, int16le, uint16le, int16be, uint16be, cfloat32, cfloat32le, cfloat32be, cfloat64, cfloat64le, cfloat64be, int8, uint8, bit')
    options.add_argument('-fslbvec',metavar=('<bvecs>'),help='specify bvec path if path is different from the path to the dwi or the file has an unusual extention')
    options.add_argument('-fslbval',metavar=('<bvals>'),help='specify bval path if path is different from the path to the dwi or the file has an unusual extention')
    options.add_argument('-bids',metavar=('<bids>'),help='specify bids.json path if path is different from the path to the dwi or the file has an unusual extention')
    options.add_argument('-n_cores',metavar=('<ncores>'),help='specify the number of cores to use in parallel tasts, by default designer will use available cores - 2', default=-3)
    options.add_argument('-echo_time',metavar=('<TE1,TE2,...>'),help='specify the echo time used in the acquisition (comma separated list the same length as number of inputs)')
    options.add_argument('-bshape',metavar=('<beta1,beta2,...>'),help='specify the b-shape used in the acquisition (comma separated list the same length as number of inputs)')

    dki_options = cmdline.add_argument_group('tensor options for the TMI script')
    dki_options.add_argument('-DKI', action='store_true', help='Include DKI parameters in output folder (mk,ak,rk)')
    dki_options.add_argument('-DTI', action='store_true', help='Include DTI parameters in output folder (md,ad,rd,fa,eigenvalues, eigenvectors')
    dki_options.add_argument('-WDKI', action='store_true',help='output the kurtosis tensor with W cumulant rather than K')    
    dki_options.add_argument('-akc_outliers', action='store_true', help='brute force K tensor outlier rejection')
    dki_options.add_argument('-fit_constraints',metavar=('<string>'),help='constrain the wlls fit (default 0,1,0)')
    dki_options.add_argument('-fit_smoothing',metavar=('<percentile>'),help='NLM smoothing on wlls fit')

    smi_options = cmdline.add_argument_group('tensor options for the TMI script')
    smi_options.add_argument('-SMI', action='store_true',help='Perform estimation of SMI (standard model of Diffusion in White Matter). Please use in conjunction with the -bshape, -echo_time, -sigma, and -compartments options.')  
    smi_options.add_argument('-compartments', metavar=('<compartments>'),help='SMI compartments (IAS, EAS, and FW), default=IAS,EAS')  
    smi_options.add_argument('-sigma', metavar=('<noisemap>'),help='path to noise map for SMI parameter estimation')

    wmti_options = cmdline.add_argument_group('tensor options for the TMI script')
    wmti_options.add_argument('-WMTI', action='store_true', help='Include WMTI parameters in output folder (awf,ias_params,eas_params)')


def execute(): #pylint: disable=unused-variable
    from mrtrix3 import app, path, run, MRtrixError #pylint: disable=no-name-in-module, import-outside-toplevel
    import lib.tensor as tensor
    import numpy as np
    from ants import image_read
    import pandas as pd

    app.make_scratch_dir()

    dwi_metadata = get_input_info(
        app.ARGS.input, app.ARGS.fslbval, app.ARGS.fslbvec, app.ARGS.bids)

    assert_inputs(dwi_metadata, None, None)

    convert_input_data(dwi_metadata)

    shell_table = create_shell_table(dwi_metadata)
    shell_rows = ['b-value', 'b-shape', 'n volumes', 'echo time']
    shell_df = pd.DataFrame(data = shell_table, 
                  index = shell_rows)
    print('input DWI data has properties:')
    print(shell_df)

    app.goto_scratch_dir()

    run.command('mrconvert dwi.mif -export_grad_fsl dwi.bvec dwi.bval dwi.nii', show=False)
    nii = image_read('dwi.nii')
    dwi = nii.numpy()

    outdir = path.from_user(app.ARGS.output, True)
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    nii = image_read('dwi.nii')
    dwi = nii.numpy()
    bvec = np.loadtxt('dwi.bvec')
    bval = np.loadtxt('dwi.bval')

    order = np.floor(np.log(abs(np.max(bval)+1)) / np.log(10))
    if order >= 2:
        bval = bval / 1000

    maxb = 3.2
    dwi = dwi[:,:,:,bval < maxb]
    bvec = bvec[:,bval < maxb]
    bval = bval[bval < maxb]

    if app.ARGS.mask:
        mask = image_read(path.from_user(app.ARGS.mask)).numpy()
    else:
        mask = np.ones(dwi.shape[:-1])
    mask = mask.astype(bool)

    if app.ARGS.fit_constraints:
        constraints = app.ARGS.fit_constraints
        if type(constraints) == str:
                constraints = constraints.split(",")
                constraints = [int(i) for i in constraints]
        else:
            raise MRtrixError("Constraints must be a 3 element comma separated string (i.e. 0,1,0)")
    else:
        constraints = [0,0,0]

    if app.ARGS.DTI:
        print('...Single shell DTI fit...')
        dtishell = (bval <= 0.1) | ((bval > .5) & (bval <= 1.5))
        dwi_dti = dwi[:,:,:,dtishell]
        bval_dti = bval[dtishell]
        bvec_dti = bvec[:,dtishell]
        grad_dti = np.hstack((bvec_dti.T, bval_dti[None,...].T))
        dti = tensor.TensorFitting(grad_dti, int(app.ARGS.n_cores))
        dt_dti, s0_dti, b_dti = dti.dti_fit(dwi_dti, mask)

    if app.ARGS.DKI or app.ARGS.WDKI:
        print('...Multi shell DKI fit with constraints = ' + str(constraints))
        grad = np.hstack((bvec.T, bval[None,...].T))
        dki = tensor.TensorFitting(grad, int(app.ARGS.n_cores))
        dt_dki, s0_dki, b_dki = dki.dki_fit(dwi, mask, constraints=constraints)

    if app.ARGS.akc_outliers:
        from lib.mpunits import vectorize
        import scipy.io as sio

        dwd = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        mat = sio.loadmat(os.path.join(dwd,'constant','dirs10000.mat'))
        dir = mat['dir']

        print('...Outlier detection...')
        if not app.ARGS.DKI:
            raise MRtrixError("AKC Outlier detection must be accompanied by DKI option")
        else:
            akc_mask = dki.outlierdetection(dt_dki, mask, dir)
            
        akc_mask = vectorize(akc_mask, mask).astype(bool)
        print('N outliers = %s' % (np.sum(akc_mask)))

        dwi_new = refit_or_smooth(akc_mask, dwi, n_cores=int(app.ARGS.n_cores))
        dt_new,_,_ = dki.dki_fit(dwi_new, akc_mask)

        x,y,z = np.where(akc_mask == 1)
        DT = vectorize(dt_dki, mask)
        DT[x,y,z,:] = dt_new.T
        dt_dki = vectorize(DT, mask)
        akc_mask = dki.outlierdetection(dt_dki, mask, dir)
        akc_mask = vectorize(akc_mask, mask).astype(bool)
        print('N outliers = %s' % (np.sum(akc_mask)))
    else:
        akc_mask = np.zeros_like(mask)

    if app.ARGS.fit_smoothing:
        print('...Nonlocal smoothing...')
        dwi_new = refit_or_smooth(akc_mask, dwi, mask=mask, smoothlevel=int(app.ARGS.fit_smoothing))
        if app.ARGS.DTI:
            dt_dti,_,_ = dti.dti_fit(dwi_new[:,:,:,dtishell], mask)
        if (app.ARGS.DKI or app.ARGS.WDKI):
            dt_dki,_,_ = dki.dki_fit(dwi_new, mask)

    if app.ARGS.DTI:
        print('...extracting and saving DTI maps...')
        params_dti = dti.extract_parameters(dt_dti, b_dti, mask, extract_dti=True, extract_dki=False, fit_w=False)
        save_params(params_dti, nii, model='dti', outdir=outdir)

    if app.ARGS.DKI:
        print('...extracting and saving DKI maps...')
        params_dki = dki.extract_parameters(dt_dki, b_dki, mask, extract_dti=True, extract_dki=True, fit_w=False)
        save_params(params_dki, nii, model='dki', outdir=outdir)

    if app.ARGS.WDKI:
        print('...extracting and saving WDKI maps...')
        params_dwi = dki.extract_parameters(dt_dki, b_dki, mask, extract_dti=False, extract_dki=True, fit_w=True)
        save_params(params_dwi, nii, model='wdki', outdir=outdir)

    if app.ARGS.WMTI:
        import dipy.reconst.dki as dki
        import dipy.reconst.dki_micro as dki_micro
        from dipy.core.gradients import gradient_table

        print('...WMTI fit...')
        gtab = gradient_table(bval*1000, bvec)
        dki_micro_model = dki_micro.KurtosisMicrostructureModel(gtab)
        dki_micro_fit = dki_micro_model.fit(dwi, mask=mask)
        
        params_wmti = {}
        params_wmti['awf'] = dki_micro_fit.awf
        params_wmti['tort'] = dki_micro_fit.tortuosity
        params_wmti['had'] = dki_micro_fit.hindered_ad
        params_wmti['hrd'] = dki_micro_fit.hindered_rd
        params_wmti['axd'] = dki_micro_fit.axonal_diffusivity
        save_params(params_wmti, nii, model='wmti', outdir=outdir)

    if app.ARGS.SMI:
        from lib.smi import SMI
        if not app.ARGS.sigma:
            print('WARNING: SMI is poorly conditioned without prior estimate of sigma')
            sigma = None
        else:
            sigma = image_read(path.from_user(app.ARGS.sigma)).numpy()

        if app.ARGS.compartments:
            compartments = app.ARGS.compartments
            if type(compartments) == str:
                compartments = compartments.split(",")
                compartments = [str(i) for i in compartments]
            else:
                raise MRtrixError(" Compartments must be a comma sepearated string (i.e. IAS,EAS)")
        else: 
            compartments = ['IAS', 'EAS']

        print('...SMI fit...')
        smi = SMI(bval=bval, bvec=bvec)
        smi.set_compartments(compartments)
        smi.set_echotime(dwi_metadata['echo_time'])
        smi.set_bshape(dwi_metadata['bshape'])
        params_smi = smi.fit(dwi, mask=mask, sigma=sigma)
        save_params(params_smi, nii, model='smi', outdir=outdir)

def main():
    import mrtrix3
    mrtrix3.execute() #pylint: disable=no-member
